
%{

/* Copyright (c) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
 * 2005, 2006, 2007, 2008 by Arkkra Enterprises */
/* All rights reserved */

/* parser for music publication program. Uses lex to get tokens. */


#include "rational.h"
#include "globals.h"
#include "defines.h"
#include "structs.h"

#define YYDEBUG 1


/* how many items to allocate at a time for
 * "with" list, and location lists, etc */
#define ITEMS	(4)


static struct MAINLL *Currstruct_p;	/* main list struct being filled in */
static struct BLOCKHEAD *Currblock_p;	/* current block, if non-zero */

static struct INPCOORD *Curr_loc_info_p; /* location info being collected */
static struct USER_SYMBOL *Curr_usym_p;	/* user-defined symbol being defined */


static short Gotheader, Gotfooter;	/* boolean: to make sure each is defined only once */
static short Gothead2, Gotfoot2;
static short Getting_tup_dur;		/* YES if currently collection tuplet
					 * duration rather than GRPSYL dur */

static struct GRPSYL *Curr_grpsyl_p;	/* current GRPSYL struct being filled in */
static struct GRPSYL *Last_grpsyl_p;	/* previous GRPSYL struct filled in (contains
					 * defaults for the current one */
static struct TIMELIST *Extra_time_p;	/* If user enters times to be added
					 * together, like 2.+16 this points to
					 * a linked list of the extra times */
static struct TIMELIST *Curr_timelist_p;/* current additive time, used to know
					 * where to link items to the list */
static struct TIMELIST *Last_alloced_timelist_p; /* Most recently allocated
					 * list. Keeping track of this lets us
					 * avoid trying to free the list
					 * when Extra_time_p is pointing
					 * to a list from an SSV. */
static struct GRPSYL *Prev_grpsyl_p;	/* Like Last_grpsyl_p except in the
					 * case of additive times,
					 * in which case it points to the
					 * first group. This is needed in case
					 * the following group is to get its
					 * time value by defaulting to that
					 * of its previous group. */
static int Extra_basictime;		/* for saving basictime, when doing
					 * additive times */
static int Doing_timeunit = NO;		/* YES if gathering timeunit param */
static struct GRPSYL *Lyrics_p;		/* pointer to current list of lyrics */
static char **Curr_marklist;		/* current "with" list */
static struct PRINTDATA **Next_print_link_p_p;	/* points to where to attach next
						 * print command that we get */
static int Item_count;			/* items in current list */
static int Max_items;			/* current amount of space in lists
					 * "with" list, curve locations, etc) */

static int Plus_minus;			/* if 1 add, if -1 subtract offset */
static int Defining_multiple = NO;	/* if defining multiple voices at once */
static int Chord_at_a_time = NO;	/* YES for chord-at-a-time,
					 * NO for voice-at-a-time */
static int EM_staff;			/* Staff number of most recent
					 * emptymeas parameter usage */
static short Endingloc;			/* if current bar starts or ends an
					 * ending, etc (STARTITEM, ENDITEM, etc) */
static short Got_ending;		/* YES if got ending on this bar */

static int Til_bars;			/* number of bars in "til" clause */
static double Dist;			/* dist for a specific STUFF */
static int Dist_usage;			/* meaning of Dist field */
static int Aligntag = NOALIGNTAG;	/* align tag number for STUFF */
static short User_meas_time;		/* YES if user specified a time for
					 * an 'm' item. Only valid on mr */
static int Order_place;			/* PL_* for markorder in SSV */
static int Order_prio;			/* value in markorder in SSV */
static int Gracebackup;			/* how many grace notes to back up
					 * before starting a "stuff" */
static float Til_offset;		/* beat count in "til" clause */
static int Linetype;			/* L_DOTTED, L_DASHED, or L_NORMAL */

static char *String1, *String2;		/* to temporarily keep track
					 * of labels and other temporary
					 * strings */
static char Tmpbuff[20];		/* temporary buffer for converting
					 * float numbers */
static char Stringbuff[100];		/* temporary string storage */
static int Titlefont, Titlesize;	/* font/size for title command */
static double Extra;			/* extra vertical space for prints */
static short Good_till_canceled = NO;	/* if [] things should be continued
					 * until turned off */
static short Curr_paratype = J_JUSTPARA; /* current paragraph type */
static struct TIMEDSSV *Curr_tssv_p;	/* current mid-measure param change */
static char Timerep[MAXTSLEN];		/* Internal representation of time
					 * signature being collected */
static int Tsig_offset;			/* Current place in Timerep while
					 * collecting a time signature */
static char *Alt_timesig_list;		/* If user specified alternating
					 * time signatures, this points to
					 * the beginning of the list, for
					 * when we need to wrap around */
static char *Next_alt_timesig;		/* If user specified alternating
					 * time signatures, this points to
					 * the one for the next measure */
static short Tsig_visibility;		/* If EACHMEAS, alternating
					 * time signatures will be done
					 * by explicitly printing
					 * the appropriate time signature at
					 * the beginning of every measure,
					 * rather than printing them all just
					 * once at the beginning of the run
					 * of measures that alternate. */

/* Row index values must match the PS_* value definitions in defines.h.
 * Column [0] is height in portrait mode, and column [1] is height. */
double pagesztbl[7][2] = {
	{ 8.5, 11.0 },		/* PS_LETTER */
	{ 8.5, 14.0 },		/* PS_LEGAL */
	{ 8.5, 13.0 },		/* PS_FLSA */
	{ 5.5, 8.5 },		/* PS_HALFLETTER */
	{ 8.26, 11.69 },	/* PS_A4 */
	{ 5.84, 8.26 },		/* PS_A5 */
	{ 4.12, 5.84 }		/* PS_A6 */
};



/* static functions */
static char *stripquotes P((char *string));	/* rmv double quotes
						 * from string */
static void chkdup_headfoot P((int flag, char *which));
static void end_prev_context P((void));
static void clone_staff_ssvs P((struct MAINLL *insert_p));
static void clone_voice_ssvs P((struct MAINLL *insert_p));
static void clone1ssv P((struct SSV *clone_src_p, struct MAINLL *insert_p,
		int staffno, int voiceno));
static void chk_ssv_ranges P((int context));
static void var_valid P((void));	 /* check if okay to set location var */
static void proc_printcmd P((int justifytype, struct INPCOORD *inpc_p,
		char *str, int font, int size, int got_nl,
		int isPostScript, double extra));
static double extra_needed P((int font, int size, char *string));
static void keyword_notes P((char *str));
static void free_extra_time P((void));
static void tsig_item P((int item));


%}



%union
{
int	intval;		/* for tokens/nonterminals that return int values */
char	*stringval;	/* for tokens/nonterminals that return strings */
float	floatval;	/* for nonterminals that return floats */
RATIONAL ratval;	/* for nonterminals that return rational numbers */
struct INPCOORD *inpcoord_p;	/* for nonterminals that return coord info */
}




%token <intval> T_SCORE T_STAFF T_VOICE T_GRIDS
%token <intval> T_BLOCKHEAD
%token <intval> T_NUMBER T_EQUAL T_SEMICOLON T_NEWLINE T_COLON T_COMMA
%token <intval> T_LBRACKET T_RBRACKET T_LBRACE T_RBRACE
%token <intval> T_DASH T_PLUS T_PLACE T_ALL T_WITH T_CUE T_GRACE T_XNOTE T_DIAM
%token <intval> T_DOT T_MULTIWHOLE T_LET_M T_LET_R T_LET_S T_LET_U
%token <intval> T_PITCH T_SHARP T_AMPERSAND T_LET_N T_LET_X T_DBLFLAT
%token <intval> T_LETTER T_LYRICS T_PAD T_LPAREN T_RPAREN T_LVAR
%token <intval> T_MUSIC
%token <intval> T_HEADSHAPES
%token <intval> T_HS
%token <intval> T_SWINGUNIT
%token <intval> T_POSTSCRIPT
%token <intval> T_SCOREFEED T_BARTYPE T_HAT T_TILDE T_QUESTION
%token <intval> T_EM_BEGIN T_EM_END
%token <intval> T_LINETYPE T_LINE T_TO T_STAR T_BULGE
%token <intval> T_CURVE T_TIE T_PRINTTYPE T_NL T_STAFFLINES T_FFAMILY
%token <intval> T_PARAGRAPH T_PARATYPE
%token <intval> T_NUMVAR T_FNUMVAR T_2FNUMVAR
%token <intval> T_SHORTEN
%token <intval> T_RATNUMLISTVAR T_RANGELISTVAR T_TIMEUNIT
%token <intval> T_FONTVAR T_STRVAR T_CLEFVAR T_VVAR T_BARSTLISTVAR T_SCORESEP
%token <intval> T_FONT T_SLASH T_KEY T_TIME T_CUT T_COMMON T_CLEF T_MULTIREST
%token <intval> T_YESNOVAR T_BM T_ESBM T_EBM T_OCTAVE T_ROLL
%token <intval> T_VISVAR T_WHEREUSED T_UNSET 
%token <intval> T_USEACCS T_NONE T_NONNAT
%token <intval> T_SYMBOL T_BBOX T_STEMOFFSET
%token <intval> T_VCOMBINE T_VCOMBVAL
%token <intval> T_SCOREPAD
%token <intval> T_UP T_DOWN T_ENDING T_ENDENDING T_FAMILY
%token <intval> T_ENDSTYLE T_ENDTYPE T_SLASHMARK T_ALT T_TITLE
%token <intval> T_REHEARSAL T_NUM T_LET T_MNUM T_TRANSPOSE T_XPOS_INT
%token <intval> T_PEDAL T_PHRASE T_MUSSYM T_L_ANGLE T_MODIFIER T_R_ANGLE T_TIL
%token <intval> T_MIDI T_PEDSTYLE T_PEDSTAR
%token <intval> T_HIDECHANGES T_SLUR T_UNITS T_UNITTYPE
%token <intval> T_TAB T_TICKS T_NOWHERE_SLIDE
%token <intval> T_LEN T_EXCLAM
%token <intval> T_OTHERTEXT T_ORDER
%token T_DRUM T_RPT
%token <stringval> T_STRING
%token T_HO
%token <intval> T_REH_STYLE T_REHTYPE
%token <intval> T_L_DBLANGLE T_R_DBLANGLE
%token <intval> T_SLOPE
%token <intval> T_PSVAR T_PAGESIZE T_ORIENTATION
%token <intval> T_ALIGN

%type <intval> num vschemetype staffnum voicenum
%type <intval> opt_n flagval dots
%type <intval> rangetail grp_staff_range staff_range verse_range
%type <intval> opt_minus pitch octave minuslist pluslist plus_or_minus
%type <intval> opt_accidental opt_octave dir opt_dir opt_side acc_symbol
%type <intval> crossbeam printtup
%type <intval> value
%type <floatval> steps_offset
%type <floatval> stuff_dist
%type <intval> opt_ticks tab_string_list opt_str_acc
%type <stringval> opt_bend
%type <intval> east_or_west north_or_south bar_opt
%type <intval> printtype notelist some_notes opt_other_staff
%type <intval> linetype pedstyle font_family title_font
%type <intval> updown transpose_interval opt_loc
%type <intval> marktype orderitem
%type <intval> opt_rs
%type <intval> voice_spec
%type <intval> stuff_type opt_size opt_modifier opt_place til_suffix majmin
%type <intval> opt_plus_or_minus hor_offset
%type <intval> paramname
%type <intval> end_type
%type <intval> opt_paratype
%type <intval> precbartype
%type <intval> barline
%type <intval> bartype
%type <intval> opt_ts_visibility
%type <intval> pagesize orientation
%type <intval> opt_vcombine_qualifier
%type <intval> useaccs_value useaccs_qual
%type <stringval> string glabel mark_item opt_decimal_part numstr
%type <stringval> loc_variable variable shorthand_mark ending_info
%type <stringval> rehearsal_mark opt_string
%type <stringval> opt_line_str
%type <floatval> floatnum yoffsetval padspec beat_offset
%type <ratval> time_item basic_time_val opt_time
%type <ratval> swing_time
%type <inpcoord_p> location alloc_loc

%expect 4

%%


song:	item

	|
	song item
	;

item:	context
	
	|
	assignment

	|
	title

	|
	paragraph

	|
	music_input

	|
	directive T_NEWLINE

	|
	stringpair

	|
	os_directive opt_semi T_NEWLINE

	|
	T_NEWLINE
	{
		/* for blank lines -- nothing to do */
	}

	|
	error T_NEWLINE
	{
		/* to keep going after stumbling on an input error */
		Getting_tup_dur = NO;
		Good_till_canceled = NO;
		Defining_multiple = NO;
		Curr_grpsyl_p = (struct GRPSYL *) 0;
		Last_grpsyl_p = (struct GRPSYL *) 0;
		Prev_grpsyl_p = (struct GRPSYL *) 0;
	};

context:	ssv_context
	{
		/* fill in context field of the SSV struct */
		Currstruct_p->u.ssv_p->context = (short) Context;
	}

	|
	blockhead_context
	{
		Curr_family = Score.fontfamily;
	}

	|
	grid_context
	{}

	|
	headshape_context
	{}


	|
	symbol_context
	{}

	|
	music_context new_context
	{
		/* have lots of separate structs in music context, so
		 * mark that we have no current struct */
		Currstruct_p = (struct MAINLL *) 0;

		/* make sure at least 1 staff is visible. Not good to
		 * check when user sets visible, because it would be
		 * nice to first set everything invisible and later
		 * turn one on. So if we wait till now to check,
		 * user doesn't have to switch back and forth from
		 * score to staff to get 1 voice visible. */
		check_at_least1visible();
	};

ssv_context:	T_SCORE new_context alloc_ssv
	{
		Context = C_SCORE;
	}

	|
	T_STAFF new_context { begin_range(PL_UNKNOWN); } ssv_stafflist alloc_ssv
	{
		/* Make sure all the staffs are in range */
		chk_ssv_ranges(C_STAFF);

		/* Set current SSV's staff to the first one on the list.
		 * If there is more than one, we will clone the SSV later.
		 * Note that it is possible that the staff number the user
		 * specified is actually illegal, but we will use it anyway,
		 * because it keeps the code simpler to pretend all is well,
		 * and ssv.c code will ignore such illegal SSVs anyway.
		 */
		if (Svrangelist_p != 0 && Svrangelist_p->stafflist_p != 0) {
			Currstruct_p->u.ssv_p->staffno
					= Svrangelist_p->stafflist_p->begin;
		}
		Context = C_STAFF;
	}

	|
	T_VOICE new_context { begin_range(PL_UNKNOWN); } ssv_sv_list alloc_ssv
	{
		/* Make sure all the staffs and voices are in range */
		chk_ssv_ranges(C_VOICE);

		/* Set current SSV's staff/voice to the first one on the list.
		 * If there are more than, we will clone the SSV later */
		if (Svrangelist_p != 0 && Svrangelist_p->stafflist_p != 0) {
			Currstruct_p->u.ssv_p->staffno
					= Svrangelist_p->stafflist_p->begin;
		}
		if (Svrangelist_p != 0 && Svrangelist_p->vnolist_p != 0) {
			Currstruct_p->u.ssv_p->voiceno
					= Svrangelist_p->vnolist_p->begin;
		}
		Context = C_VOICE;
	};

ssv_stafflist: ssv_staff_item
	|
	ssv_stafflist sv_and ssv_staff_item
	;

ssv_staff_item:
	grp_stafflist
	{
		add_to_sv_list();
	};

ssv_sv_list: ssv_sv_item
	|
	ssv_sv_list sv_and ssv_sv_item
	;

ssv_sv_item: grp_stafflist voice_spec
	{
		add_to_sv_list();
	};

alloc_ssv: 
	{
		/* allocate an SSV struct */
		Currstruct_p = newMAINLLstruct(S_SSV, yylineno);
	};

new_context:
	{
		/* entering new context, if we have a current struct,
		 * we were in a different context, so add its info to the list */
		end_prev_context();
	};

blockhead_context:	T_BLOCKHEAD
	{
		struct FEED *feed_p;		/* where to attach blocks */
		int calloced;			/* YES if dynamically alloced */

		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = $1;
		calloced = NO;

		switch ($1) {
		case C_HEADER:
			/* We only allow one of these */
			chkdup_headfoot(Gotheader++, "header");
			Currblock_p = &Header;
			break;
		case C_HEAD2:
			chkdup_headfoot(Gothead2++, "header2");
			Currblock_p = &Header2;
			break;
		case C_FOOTER:
			chkdup_headfoot(Gotfooter++, "footer");
			Currblock_p = &Footer;
			break;
		case C_FOOT2:
			chkdup_headfoot(Gotfoot2++, "foot2");
			Currblock_p = &Footer2;
			break;
			break;
		case C_TOP:
		case C_TOP2:
		case C_BOT:
		case C_BOT2:
		case C_BLOCK:
			CALLOC(BLOCKHEAD, Currblock_p, 1);
			calloced = YES;
			break;
		default:
			pfatal("Unknown block-type context");
			/*NOTREACHED*/
			break;
		}
		set_win_coord(Currblock_p->c);
		/* Remember where to start attaching "print" commands */
		Next_print_link_p_p = &(Currblock_p->printdata_p);

		/* The dynamically allocated blocks
		 * (not the static head/foot blocks)
		 * need extra processing to populate
		 * the main list properly with FEEDs.
		 */
		if (calloced == YES) {
			if (Mainlltc_p == 0 || Mainlltc_p->str != S_FEED) {
				/* User had not specified a feed
				 * right before this block, so we need
				 * to add one for them implicitly.
				 * Create a new FEED and add it
				 * to the end of the main list.
				 */
				insertMAINLL(newMAINLLstruct(S_FEED, yylineno),
							Mainlltc_p);
			}
			feed_p = Mainlltc_p->u.feed_p;

			/* Normally, top/top2/bot/bot2 result in a pagefeed.
			 * But we have a special case for backward
			 * compatibility and for flexibility.
			 * User can get a separate
			 * "title" page if they really want to,
			 * by using an explicit newpage at the beginning.
			 * But in the normal case, they'd want to declare
			 * top/bot that will go on the first page without
			 * that causing a newpage. So if there hasn't been
			 * any music or block yet, we won't set newpage=YES
			 * for top/top2/bot/bot2, but otherwise we will.
			 */
			if (Context & (C_TOP|C_TOP2|C_BOT|C_BOT2)) {
				struct MAINLL *m_p;
				for (m_p = Mainlltc_p->prev; m_p != 0;
							m_p = m_p->prev) {
					if (m_p->str == S_BAR ||
							m_p->str == S_BLOCKHEAD) {
						/* There was music or block,
						 * so this is not
						 * at very beginning,
						 * so no special case:
						 * there will be new page.
						 */
						feed_p->pagefeed = YES;
						break;
					}
				}
			}

			/* Now check for redefined blocks and
			 * attach the BLOCKHEAD onto the FEED or the
			 * main list, as appropriate.
			 */
			switch (Context) {
			case C_TOP:
				if (feed_p->top_p != 0) {
					warning("top redefined before being used");
				}
				feed_p->top_p = Currblock_p;
				break;
			case C_TOP2:
				if (feed_p->top2_p != 0) {
					warning("top2 redefined before being used");
				}
				feed_p->top2_p = Currblock_p;
				break;
			case C_BOT:
				if (feed_p->bot_p != 0) {
					warning("bot redefined before being used");
				}
				feed_p->bot_p = Currblock_p;
				break;
			case C_BOT2:
				if (feed_p->bot2_p != 0) {
					warning("bot2 redefined before being used");
				}
				feed_p->bot2_p = Currblock_p;
				break;
			case C_BLOCK:
				insertMAINLL(newMAINLLstruct(S_BLOCKHEAD, yylineno),
						Mainlltc_p);
				Mainlltc_p->u.blockhead_p = Currblock_p;
				/* These blocks count like "music" for purposes
				 * of things that can only be set
				 * before any "music"
				 */
				Got_some_data = YES;
				break;
			case C_HEADER:
			case C_HEAD2:
			case C_FOOTER:
			case C_FOOT2:
				/* These are static, not in main list,
				 * so nothing needs to be done with them.
				 */
				break;
			default:
				pfatal("unexpected context (0x%x) for block", Context);
				/*NOTREACHED*/
				break;
			}

			/* We need a feed after a block.
			 * Use -1 as lineno to mark it as internally generated.
			 * That way if user puts their own explicit feed
			 * next in the input, we can know we can
			 * discard this internally generated one.
			 */
			if (Context & C_BLOCK) {
				insertMAINLL(newMAINLLstruct(S_FEED, -1), Mainlltc_p);
			}
		}
	};

music_context:	T_MUSIC
	{
		Context = C_MUSIC;
	};

grid_context:	T_GRIDS
	{
		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = C_GRIDS;
	};

stringpair:
	string string T_NEWLINE
	{
		if (Context == C_GRIDS) {
			add_grid($1, $2);
		}
		else if (Context == C_HEADSHAPES) {
			/* skip past the font/size bytes */
			add_shape($1 + 2, $2 + 2);
		}
		else {
			l_warning(Curr_filename, yylineno - 1,
				"unexpected text strings (possibly intended to be in grids or headshapes context, or missing + for concatenation?)");
		}
	};

headshape_context:
	T_HEADSHAPES
	{
		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = C_HEADSHAPES;
	}
	;

symbol_context:
	T_SYMBOL string
	{
		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = C_SYMBOL;
		/* The +2 to skip font/size */
		Curr_usym_p = alloc_usym($2 + 2);
	}
	;

staffnum:	num
	;

voicenum:	num
	{
		$$ = $1;
		/* later on we check for being within current range */
	};

num:	T_NUMBER
	{
		$$ = atoi(yytext);
	};

value:	opt_minus num
	{
		$$ = $1 * $2;
	}
	;

assignment:	assign terminator
	;

terminator:	T_SEMICOLON

	|
	T_NEWLINE
	;

assign:	T_NUMVAR T_EQUAL opt_minus num
	{
		/* Only a few things are allowed to have a negative number.
		 * Other code assumes that that restriction is enforced here,
		 * so we only honor the minus if it is legal. */
		if ($3 == -1) {
			if ($1 != STAFFPAD && $1 != SYLPOSITION) {
				yyerror("negative value not allowed");
			}
			else {
				assign_int($1, -($4), Currstruct_p);
			}
		}
		else {
			assign_int($1, $4, Currstruct_p);
		}
	}

	|
	T_NUMVAR T_EQUAL
	{
		/* only restcombine and gridfret can be set to empty */
		if ($1 == RESTCOMBINE) {
			assign_int($1, NORESTCOMBINE, Currstruct_p);
		}
		else if ($1 == GRIDFRET) {
			assign_int($1, NOGRIDFRET, Currstruct_p);
		}
		else {
			yyerror("parameter value required");
		}
	}

	|
	T_FNUMVAR T_EQUAL floatnum
	{
		assign_float($1, $3, Currstruct_p);
	}

	|
	T_2FNUMVAR T_EQUAL floatnum T_COMMA floatnum
	{
		assign_2floats($1, $3, $5, Currstruct_p);
	}

	|
	T_SHORTEN T_EQUAL floatnum shorten_opts
	{
		assign_float($1, $3, Currstruct_p);
	}

	|
	T_PSVAR T_EQUAL pagesize orientation
	{
		double multiplier;
		multiplier = (Score.units == INCHES ? 1.0 : CMPERINCH);
		assign_float(PAGEWIDTH, pagesztbl[$3][$4] * multiplier, Currstruct_p);
		assign_float(PAGEHEIGHT, pagesztbl[$3][$4 ^ 1] * multiplier, Currstruct_p);
	}

	|
	T_VCOMBINE T_EQUAL { begin_range(PL_UNKNOWN);  } vcombine_value
	{ }

	|
	T_PAD T_EQUAL opt_minus floatnum
	{
		/* specified in stepsizes, stored in inches */
		assign_float(PAD, $3 * $4 * STEPSIZE, Currstruct_p);
	}

	|
	T_VVAR T_EQUAL num vschemetype
	{
		assign_vscheme ($3, $4, Currstruct_p);
	}

	|
	T_RATNUMLISTVAR T_EQUAL allocbeamlist ratnumlist opt_rs
	{
		set_beamlist(Currstruct_p);
		if (Currstruct_p != 0 && Currstruct_p->u.ssv_p != 0) {
			Currstruct_p->u.ssv_p->beamrests = ($5 & 0x1);
			Currstruct_p->u.ssv_p->beamspaces = (($5 >> 1) & 0x1);
		}
	}

	|
	T_RANGELISTVAR T_EQUAL alloc_rangelist rangelist
	{
		set_staffset($1, Currstruct_p);
	}

	|
	T_BARSTLISTVAR T_EQUAL alloc_barstlist barstlist
	{
		set_barstlist(Currstruct_p);
	}

	|
	mark_order T_EQUAL orderlist
	{
		if (Currstruct_p != 0) {
			chk_order(Currstruct_p->u.ssv_p, Order_place);
		}
	}

	|
	T_KEY T_EQUAL keyspec

	|
	T_USEACCS T_EQUAL useaccs_value
	{
		assign_int($1, $3, Currstruct_p);
	}

	|
	T_TIME T_EQUAL timesig
	{
		/* all the work for this is done in timesig */
	}

	|
	T_STAFFLINES T_EQUAL stafflinedef
	{
	}

	|
	T_FONTVAR T_EQUAL T_FONT
	{
		set_font($1, $3, Currstruct_p);
	}

	|
	T_FAMILY T_EQUAL T_FFAMILY
	{
		set_font($1, $3, Currstruct_p);
	}

	|
	T_CLEFVAR T_EQUAL T_CLEF
	{
		if (contextcheck(C_SCORE|C_STAFF,"clef parameter") == YES) {
			used_check(Currstruct_p, CLEF, "clef");
			if (Context == C_STAFF && has_tab_staff()) {
				/* We used to be able to make this just a
				 * warning, but now that is hard, because
				 * we'd want to let it assign for any
				 * non-tab, but it is hard to know at this
				 * point which SSVs will be tab and which					 * non-tab, because we may not have created
				 * them all yet. We cannot let the tab ones
				 * get assigned, because later code would
				 * get confused by clef not being TABCLEF.
				 * While we could figure this all out with
				 * enough code, it really is a user error,
				 * so not worth bending over backwards
				 * to try to keep it just a warning.
				 */
				l_yyerror(Curr_filename, yylineno,
					"can't set clef on tab staff");
			}
			Currstruct_p->u.ssv_p->clef = $3;
			Currstruct_p->u.ssv_p->used[CLEF] = YES;
			asgnssv(Currstruct_p->u.ssv_p);
		}
	}

	|
	T_YESNOVAR T_EQUAL flagval
	{
		assign_int($1, $3, Currstruct_p);
	}

	|
	T_VISVAR T_EQUAL visval

	|
	T_UNSET paramname
	{
		if (contextcheck(C_STAFF|C_VOICE, "unset") == YES) {
			if ($2 < 0 || $2 >= NUMFLDS) {
				pfatal("invalid parameter index %d", $2);
			}
			if (Currstruct_p->u.ssv_p->used[$2] == YES) {
				warning("unsetting parameter overrides previous setting");
			}
			/* Note that if user tries to unset something that
			 * can't be set in the current context, we silently
			 * ignore that. We figure that since it's already unset,
			 * it doesn't hurt anything to unset it again,
			 * and it's easier to not bother to check.
			 */

			Currstruct_p->u.ssv_p->used[$2] = UNSET;

			/* special case: scoresep controls two fields */
			if ($2 == MINSCSEP) {
				Currstruct_p->u.ssv_p->used[MAXSCSEP] = UNSET;
			}

			/* Assign, just in case there are new interactions */
			asgnssv(Currstruct_p->u.ssv_p);
		}
	}

	|
	T_TIMEUNIT T_EQUAL { Doing_timeunit = YES; } time_item
	{
		if ( contextcheck(C_SSV, "timeunit parameter") == YES ) {
			used_check(Currstruct_p, TIMEUNIT, "timeunit");
			Currstruct_p->u.ssv_p->timeunit = $4;
			Currstruct_p->u.ssv_p->timelist_p = Extra_time_p;
			Extra_time_p = 0;
			Doing_timeunit = NO;
			Currstruct_p->u.ssv_p->used[TIMEUNIT] = YES;
		}
	}

	|
	T_SWINGUNIT T_EQUAL swing_time
	{
		if ( contextcheck(C_SSV, "swingunit parameter") == YES ) {
			used_check(Currstruct_p, SWINGUNIT, "swingunit");
			Currstruct_p->u.ssv_p->swingunit = $3;
			Currstruct_p->u.ssv_p->used[SWINGUNIT] = YES;
		}
	}

	|
	T_SCORESEP T_EQUAL minscsep maxscsep
	{
	}

	|
	T_SCOREPAD T_EQUAL minscpad maxscpad
	{
	}

	|
	T_ENDSTYLE T_EQUAL end_type
	{
		if ( contextcheck(C_SCORE, "endingstyle parameter") == YES ) {
			used_check(Currstruct_p, ENDINGSTYLE, "endingstyle");
			Currstruct_p->u.ssv_p->endingstyle = $3;
			Currstruct_p->u.ssv_p->used[ENDINGSTYLE] = YES;
		}
	}

	|
	T_REH_STYLE T_EQUAL T_REHTYPE
	{
		if ($1 == REHSTYLE) {
			if (contextcheck(C_SCORE | C_STAFF, "rehstyle parameter") == YES ) {
				used_check(Currstruct_p, REHSTYLE, "rehstyle");
				Currstruct_p->u.ssv_p->rehstyle = $3;
				Currstruct_p->u.ssv_p->used[REHSTYLE] = YES;
			}
		}
		else if ($1 == MEASNUMSTYLE) {
			if (contextcheck(C_SCORE, "measnumstyle parameter") == YES) {
				used_check(Currstruct_p, MEASNUMSTYLE, "measnumstyle");
				Currstruct_p->u.ssv_p->measnumstyle = $3;
				Currstruct_p->u.ssv_p->used[MEASNUMSTYLE] = YES;
			}
		}
		
	}

	|
	T_PEDSTYLE T_EQUAL pedstyle
	{
		if ( contextcheck(C_SCORE | C_STAFF, "pedstyle parameter") == YES ) {
			used_check(Currstruct_p, PEDSTYLE, "pedstyle");
			Currstruct_p->u.ssv_p->pedstyle = $3;
			Currstruct_p->u.ssv_p->used[PEDSTYLE] = YES;
		}
	}

	|
	T_TRANSPOSE T_EQUAL updown transpose_interval num
	{
		char *trans_name;
		trans_name = ($1 == TRANSPOSITION ?
			"transpose" : "addtranspose");
		if (contextcheck(C_SCORE|C_STAFF, trans_name) == YES) {
			if (Context == C_STAFF && has_tab_staff() == YES) {
				l_warning(Curr_filename, yylineno,
					"%s not allowed on tablature staff; ignoring",
					trans_name);
			}
			used_check(Currstruct_p, $1, trans_name);
			chk_interval($4, $5);
			if ($1 == TRANSPOSITION) {
				Currstruct_p->u.ssv_p->inttype = $4;
				Currstruct_p->u.ssv_p->intnum = $3 * $5;
			}
			else {
				Currstruct_p->u.ssv_p->addinttype = $4;
				Currstruct_p->u.ssv_p->addintnum = $3 * $5;
			}
			Currstruct_p->u.ssv_p->used[$1] = YES;
			asgnssv(Currstruct_p->u.ssv_p);
		}
	}

	|
	T_UNITS T_EQUAL T_UNITTYPE
	{
		assign_unit($3, Currstruct_p);
	}

	|
	T_STRVAR T_EQUAL string
	{
		assign_string($1, $3, Currstruct_p);
	}

	|
	T_POSTSCRIPT T_EQUAL string
	{
		if (Curr_usym_p == 0) {
			l_yyerror(Curr_filename, yylineno,
				"postscript= only valid in symbol context; maybe you meant without '=' ?");
		}
		else {
			if (Curr_usym_p->flags & US_POSTSCRIPT) {
				l_warning(Curr_filename, yylineno,
					"Multiple postscript definitions given for symbol '%s'; using last",
					Curr_usym_p->name);
				FREE(Curr_usym_p->postscript);
			}
			Curr_usym_p->postscript = $3;
			Curr_usym_p->flags |= US_POSTSCRIPT;
		}
	}

	|
	T_BBOX T_EQUAL value T_COMMA value T_COMMA value T_COMMA value
	{
		if (Curr_usym_p == 0) {
			l_yyerror(Curr_filename, yylineno,
				"bbox= only valid in symbol context");
		}
		else {
			if (Curr_usym_p->flags & US_BBOX) {
				l_warning(Curr_filename, yylineno,
					"Multiple bbox definitions given for symbol %s; using last",
					Curr_usym_p->name);
			}
			Curr_usym_p->llx = $3;
			Curr_usym_p->lly = $5;
			Curr_usym_p->urx = $7;
			Curr_usym_p->ury = $9;
			Curr_usym_p->flags |= US_BBOX;
		}
	}

	|
	T_STEMOFFSET T_EQUAL value T_COMMA value
	{
		if (Curr_usym_p == 0) {
			l_yyerror(Curr_filename, yylineno,
				"stemoffset= only valid in symbol context");
		}
		else {
			if (Curr_usym_p->flags & US_STEMOFFSET) {
				l_warning(Curr_filename, yylineno,
					"Multiple stemoffset definitions given for %s; last used",
					Curr_usym_p->name);
			}
			Curr_usym_p->upstem_y = $3;
			Curr_usym_p->downstem_y = $5;
			Curr_usym_p->flags |= US_STEMOFFSET;
		}
	}
	;

visval:
	T_WHEREUSED
	{
		/* whereused is not allowed on voice, just score and staff */
		if (contextcheck(C_SCORE|C_STAFF, "visible=whereused") == YES) {
			assign_int(VISIBLE, YES, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = YES;
		}
	}

	|
	T_LETTER
	{
		/* only valid letter is y */
		if (*yytext != 'y') {
			yyerror("visible value must be y, n, or whereused");
		}
		else if (contextcheck(C_SSV, "visible parameter")
								== YES) {
			assign_int(VISIBLE, YES, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = NO;
		}
	}

	|
	T_LET_N
	{
		if (contextcheck(C_SSV, "visible parameter") == YES) {
			assign_int(VISIBLE, NO, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = NO;
		}
	};
	
vschemetype:
	{
		/* none specified */
		$$ = V_1;
	}

	|
	T_PITCH
	{
		/* only valid value here is 'f'  */
		if (*yytext == 'f') {
			$$ = V_2FREESTEM;
		}
		else {
			yyerror("invalid voice qualifier: must be f or o");
			$$ = V_1;
		}
	} 

	|
	T_LETTER
	{
		/* only valid value here is 'o' */
		if (*yytext == 'o') {
			$$ = V_2OPSTEM;
		}
		else {
			yyerror("invalid voice qualifier: must be f or o");
			$$ = (int) V_1;
		}
	};

allocbeamlist:
	{
		/* NULL token for allocating a new beamlist */
		new_beamlist();
	};

ratnumlist:
	/* empty */

	|
	nonempty_ratnumlist
	;

nonempty_ratnumlist:
	time_item
	{
		add_beamlist($1);
	}

	|
	subbeam

	|
	nonempty_ratnumlist T_COMMA time_item
	{
		add_beamlist($3);
	}

	|
	nonempty_ratnumlist T_COMMA subbeam
	;

subbeam:
	T_LPAREN { begin_subbeam(); } nonempty_ratnumlist T_RPAREN
	{
		end_subbeam();
	};

time_item: basic_time_val dots opt_more_time opt_m
	{
		$$ = calcfulltime($1, $2);

		/* if filling in a GRPSYL struct, need to fill in dots.
		 * Could also be here due to beamstyle
		 * or timeunit or location time offset,
		 * in which case the Curr_grpsyl_p will be NULL,
		 * or while getting tuplet duration. */
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
						&& Getting_tup_dur == NO) {
			Curr_grpsyl_p->dots = $2;
		}
		else {
			struct TIMELIST *timelist_p;	/* walk thru list */
			struct TIMELIST *next_p;/* save next to delete curr */
			RATIONAL totaltime;	/* sum of extra times */

			/* Start with first time value (which was calculated
			 * above and is now in $$), then add on the added
			 * times, if any. */
			totaltime = $$;
			for (timelist_p = Extra_time_p; timelist_p != 0;
						timelist_p = next_p) {
				next_p = timelist_p->next;
				totaltime = radd(totaltime, timelist_p->fulltime);
				if (Doing_timeunit == NO) {
					/* For things like swingunit
					 * or tuplet time, we just need the
					 * resulting time, not the individual
					 * time units and arithmetic. */
					FREE(timelist_p);
				}
			}
			if (Extra_time_p != 0 && LE(totaltime, Zero)) {
				yyerror("Time values must add up to more than zero");
			}
			/* Current timelist is finished; can't append any
			 * more items to it. */
			Curr_timelist_p = 0;
			if (Doing_timeunit == YES) {
				/* For timeunit, we return the initial
				 * timeunit, but also preserve
				 * Extra_time_p for putting in SSV.
				 */
				$$ = calcfulltime($1, $2);
			}
			else {
				/* the list has been freed above */
				Extra_time_p = 0;
				$$ = totaltime;
			}
		}
	}

	|
	T_LET_M
	{
		$$.n = (svpath(1, TIME))->time.n;
		$$.d = (svpath(1, TIME))->time.d;
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			/* use whole note symbol as default */
			Curr_grpsyl_p->basictime = 1;
			Curr_grpsyl_p->is_meas = YES;
			User_meas_time = NO;
		}
		else {
			yyerror("'m' is not valid here");
		}
	};

swing_time:
	/* empty */
	{
		$$ = Zero;
	}

	|
	time_item
	;

opt_rs:
	/* empty */
	{
		$$ = NO;
	}

	|
	T_LET_R T_LET_S
	{
		$$ = (YES | (YES << 1));
	}

	|
	T_LET_S T_LET_R
	{
		$$ = ((YES << 1) | YES);
	}

	|
	T_LET_S
	{
		$$ = (YES << 1);
	}

	|
	T_LET_R
	{
		$$ = YES;
	};

alloc_rangelist:
	{
		/* null token for the purpose of allocating a new rangelist
		 * to be filled in with user data */
		new_staffset();
	};

rangelist:
	/* empty */

	|
	rangeitem

	|
	rangelist T_COMMA rangeitem
	;

rangeitem:	num rangetail grplabel
	{
		/* save information about this range in the list */
		/* if only a single number, not a range, rangetail will be
		 * 0, so use num as both beginning and end of range */
		add_staffset( $1, ($2 > 0 ? $2 : $1), String1, String2);
	};

rangetail:
	{
		/* no end of range, so return 0 as flag that there was only 1 number */
		$$ = 0;
	}

	|
	T_DASH num
	{
		/* make sure isn't 0, because that it illegal and would be used
		 * later as though there were no range. Other error checking
		 * on this number will be done later. */
		if ($2 == 0) {
			yyerror("second number of range may not be zero");
		}
		$$ = $2;
	};

grplabel:
	{
		/* no labels */
		String1 = String2 = (char *) 0;
	}

	|
	T_LPAREN glabel optlabel2 T_RPAREN
	{
		/* save label */
		String1 = $2;
	};

optlabel2:
	{
		/* no second label */
		String2 = (char *) 0;
	}

	|
	T_COMMA glabel
	{
		/* save label2 for later use */
		String2 = $2;
	};

glabel:	string
	;

alloc_barstlist:
	{
		/* NULL token for allocating space for a bar style list */
		new_barstlist();
	};

barstlist:
	/* empty */

	|
	barst_item

	|
	barstlist T_COMMA barst_item
	;

barst_item:
	num rangetail
	{
		/* if only 1 number, not range, rangetail will be 0, so use
		 * same number for beginning and end */
		add_barst($1, ($2 == 0 ? $1 : $2) );
	};

mark_order:	T_ORDER
	{
		Order_place = $1;
		Order_prio = 1;
		if (Currstruct_p != 0) {
			switch ($1) {
			case PL_ABOVE:
				Currstruct_p->u.ssv_p->used[ABOVEORDER] = YES;
				break;
			case PL_BELOW:
				Currstruct_p->u.ssv_p->used[BELOWORDER] = YES;
				break;
			case PL_BETWEEN:
				Currstruct_p->u.ssv_p->used[BETWEENORDER] = YES;
				break;
			}
		}
	};

orderlist:
	orderitem
	{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][$1] = Order_prio;
		}
	}
	
	|
	orderlist T_COMMA orderitem
	{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][$3] = ++Order_prio;
		}
	}
	
	|
	orderlist T_AMPERSAND orderitem
	{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][$3] = Order_prio;
		}
	};

orderitem:	marktype
	{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.ssv_p->markorder[Order_place][$1] != 0) {
				l_yyerror(Curr_filename, yylineno,
					"order item %s specified more than once", $1);
			}
		}
		$$ = $1;
	};

marktype:
	T_MUSSYM
	{
		$$ = MK_MUSSYM;
	}

	|
	T_OCTAVE
	{
		$$ = MK_OCTAVE;
	}

	|
	T_OTHERTEXT
	{
		$$ = MK_OTHERTEXT;
	}

	|
	T_MODIFIER
	{
		switch ($1) {
		case TM_CHORD:
			$$ = MK_CHORD;
			break;
		case TM_DYN:
			$$ = MK_DYN;
			break;
		default:
			yyerror("invalid text modifier in order list");
			/* set to something valid, since $$ will be used
			 * as an array subscript. */
			$$ = MK_CHORD;
			break;
		}
	}

	|
	T_LYRICS
	{
		$$ = MK_LYRICS;
	}

	|
	T_ENDING
	{
		$$ = MK_ENDING;
	}

	|
	T_REHEARSAL
	{
		$$ = MK_REHEARSAL;
	}

	|
	T_PEDAL
	{
		$$ = MK_PEDAL;
	} ;

timesig:
	tsiglist opt_ts_visibility
	{
		tsig_item(TSR_END);
		if (Currstruct_p != 0 && Currstruct_p->u.ssv_p != 0) {
			MALLOCA(char, Currstruct_p->u.ssv_p->timerep, Tsig_offset);
			strcpy(Currstruct_p->u.ssv_p->timerep, Timerep);
			assign_timesig(Currstruct_p, $2, &Next_alt_timesig);
			Tsig_visibility = $2;
			/* If there are alternating time signatures,
			 * need to save pointer to this list */
			if (Next_alt_timesig != 0) {
				Alt_timesig_list = Timerep;
			}
			else {
				Alt_timesig_list = 0;
			}
		}
		/* Reset for next time signature usage, if any */
		Tsig_offset = 0;
	};

tsiglist:
	meas_tsig

	|
	tsiglist { tsig_item(TSR_ALTERNATING); } meas_tsig
	{
		/* this is for alternating time signature */
	};

meas_tsig:
	fraction
	{
	}

	|
	meas_tsig T_PLUS { tsig_item(TSR_ADD); } fraction
	{
		/* This would be for things like 2/4 + 3/8 */
	};

fraction:
	T_CUT
	{
		tsig_item(TSR_CUT);
	}

	|
	T_COMMON
	{
		tsig_item(TSR_COMMON);
	}

	|
	numerator T_SLASH num
	{
		tsig_item(TSR_SLASH);
		if ($3 < MINDENOMINATOR) {
			l_yyerror(Curr_filename, yylineno,
				"time signature denominator cannot be less than %d",
				MINDENOMINATOR);
			$3 = MINDENOMINATOR;
		}
		else if ($3 > MAXDENOMINATOR) {
			l_yyerror(Curr_filename, yylineno,
				"time signature denominator cannot be greater than %d",
				MAXDENOMINATOR);
			$3 = MAXDENOMINATOR;
		}
		else if (power_of2check($3, "time signature denominator") == NO) {
			$3 = 1 << drmo($3);
		}

		tsig_item($3);
	}

	|
	numerator T_PLUS T_MULTIWHOLE
	{
		/* this is something like 3+1/4 where the 1/4 get interpreted
		 * by lexer as quadwhole, but it's really 1 for numerator
		 * and 4 for denominator. 
		 */
		tsig_item(1);
		tsig_item(TSR_SLASH);
		tsig_item($3);
	}

	|
	T_MULTIWHOLE
	{
		tsig_item(1);
		tsig_item(TSR_SLASH);
		tsig_item($1);
	};

numerator:
	numerator_value

	|
	numerator T_PLUS numerator_value
	;

numerator_value: num
	{
		if ($1 < MINNUMERATOR) {
			l_yyerror(Curr_filename, yylineno,
				"time signature numerator cannot be less than %d",
				MINNUMERATOR);
			$1 = MINNUMERATOR;
		}
		else if ($1 > MAXNUMERATOR) {
			l_yyerror(Curr_filename, yylineno,
				"time signature numerator cannot be greater than %d",
				MAXNUMERATOR);
			$1 = MAXNUMERATOR;
		}
		tsig_item($1);
	};

opt_ts_visibility:
	/* empty */
	{
		$$ = PTS_ONCE;
	}

	|
	flagval
	{
		$$ = ($1 == YES ? PTS_ALWAYS : PTS_NEVER);
	};
	
opt_n:
	{
		$$ = NO;
	}

	|
	T_LET_N
	{
		/* 'n' is used in various places to mean do NOT
		 * print something that would normally be printed,
		 * so YES means we do want to inhibit the normal printing. */
		$$ = YES;
	};

keyspec: num acc_symbol majmin
	{
		if ( ($2 != '#') && ($2 != '&') ) {
			yyerror("key signature must include # or &");
		}
		else {
			assign_key($1, $2, $3, Currstruct_p);
		}
	}

	|
	pitch opt_accidental majmin
	{
		int sharps;

		/* Get number of sharps for the pitch letter */
		sharps = strchr(Circle, $1) - Circle - 1;

		/* adjust for sharp/flat */
		if ($2 == '#') {
			sharps += 7;
		}
		else if ($2 == '&') {
			sharps -= 7;
		}
		else if ($2 != 0) {
			yyerror("key letter can only be followed by # or &");
			/* set to some ridiculous value, so will fail below */
			sharps = 999;
		}

		/* Adjust if needed for minor */
		if ($3 == YES) {
			sharps -= 3;
		}

		/* disallow illegal keys, like f& */
		if (abs(sharps) > 7) {
			/* print error unless already printed one above */
			if (sharps < 100) {
				yyerror("invalid key--too many sharps or flats");
			}
		}
		else {
			assign_key(abs(sharps), (sharps >= 0 ? '#' : '&'),
							$3, Currstruct_p);
		}
	};

useaccs_value:
	flagval useaccs_qual
	{
		if ($1 == NO) {
			if ($2 != -1) {
				l_warning(Curr_filename, yylineno,
					"can't use useaccs qualifier with n; ignoring the qualifier");
			}
			$$ = UA_N;
		}
		else {
			if ($2 == -1) {
				$2 = UA_Y_NONE;
			}
			$$ = $2;
		}
	}
	;

useaccs_qual:
	{
		/* Empty means UA_Y_NONE when used with 'y', but we need
		 * to distinguish temporarily, in case user erroneously
		 * tries to use std with 'n' */
		$$ = -1;
	}

	|
	T_NONE
	{
		$$ = UA_Y_NONE;
	}

	|
	T_ALL
	{
		$$ = UA_Y_ALL;
	}

	|
	T_NONNAT
	{
		$$ = UA_Y_NONNAT;
	}
	;

stafflinedef:
	num opt_n
	{
		asgn_stafflines($1, $2 == YES ? SS_NOTHING : SS_NORMAL, Currstruct_p);
	}

	|
	num T_DRUM
	{
		asgn_stafflines($1, SS_DRUM, Currstruct_p);
	}

	|
	T_TAB
	{
		/* use default tablature */
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->strinfo = Guitar;
			asgn_stafflines(DEFTABLINES, NO, Currstruct_p);
		}
	}

	|
	T_TAB T_LPAREN tab_string_list T_RPAREN
	{
		/* user-specified tablature */
		asgn_stafflines($3, NO, Currstruct_p);
	};

tab_string_list:
	tab_item
	{
		$$ = 1;
	}

	|
	tab_string_list tab_item
	{
		/* count up the number of strings */
		$$ = $1 + 1;
	};

tab_item:	pitch opt_str_acc opt_ticks opt_octave
	{
		/* save info about this tablature string */
		if (Currstruct_p != 0) {
			add_tab_string_info($1, $2, $3, $4, Currstruct_p->u.ssv_p);
		}
	};

opt_str_acc:
	{
		$$ = 0;
	}

	|
	acc_symbol
	{
		if ($1 != '&' && $1 != '#') {
			yyerror("accidental on string can only be & or #");
		}
	};

opt_ticks:
	{
		/* no ticks */
		$$ = 0;
	}

	|
	T_TICKS
	{
		/* limit number of ticks. Has to match the number of bits
		 * used by TMP_NTICKS in grpsyl.c */
		if ($1 > MAXTICKS) {
			l_yyerror(Curr_filename, yylineno,
				"too many ' marks (%d max)", MAXTICKS);
			$1 = MAXTICKS;
		}
	};

flagval:	T_LETTER
	{
		/* only valid letter is y */
		if (*yytext != 'y') {
			yyerror("flag value must be y or n");
		}
		$$ = YES;
	}

	|
	T_LET_N
	{
		$$ = NO;
	};

minscsep:	num
	{
		if (Currstruct_p == 0) {
			l_warning(Curr_filename, yylineno,
				"wrong context for setting scoresep; ignoring");
		}
		else {
			/* Assign the value of minimum vertical space
			 * between scores. */
			assign_int(MINSCSEP, $1, Currstruct_p);
		}
	};

maxscsep:
	{
		if (Currstruct_p != 0) {
			/* if user set minimum greater than default maximum,
			 * adjust maximum to equal the minimum
			 * they specified. */
			if (Currstruct_p->u.ssv_p->minscsep
					> Currstruct_p->u.ssv_p->maxscsep) {
				Currstruct_p->u.ssv_p->maxscsep
					= Currstruct_p->u.ssv_p->minscsep;
			}
			else {
				/* fill in default */
				assign_int(MAXSCSEP, DEFMAXSCSEP, Currstruct_p);
			}
		}
	}

	|
	T_COMMA num
	{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.ssv_p->minscsep > $2) {
				yyerror("scoresep maximum smaller than minimum");
			}
			assign_int(MAXSCSEP, $2, Currstruct_p);
		}
	};

minscpad:	opt_minus num
	{
		if (Currstruct_p == 0) {
			l_warning(Curr_filename, yylineno,
				"wrong context for setting scorepad; ignoring");
		}
		else {
			/* Assign the value of minimum vertical space
			 * between scores. */
			assign_int(MINSCPAD, $1 * $2, Currstruct_p);
		}
	}
	;

maxscpad:
	{
		if (Currstruct_p != 0) {
			/* If user set minimum greater than default maximum,
			 * adjust maximum to equal the minimum
			 * they specified. */
			if (Currstruct_p->u.ssv_p->minscpad
					> Currstruct_p->u.ssv_p->maxscpad) {
				Currstruct_p->u.ssv_p->maxscpad
					= Currstruct_p->u.ssv_p->minscpad;
			}
			else {
				/* fill in default */
				assign_int(MAXSCPAD, DEFMAXSCPAD, Currstruct_p);
			}
		}
	}

	|
	T_COMMA opt_minus num
	{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.ssv_p->minscpad > $2 * $3) {
				yyerror("scorepad maximum smaller than minimum");
			}
			assign_int(MAXSCPAD, $2 * $3, Currstruct_p);
		}
	}
	;

shorten_opts:	/* empty */
	{
	}

	|
	T_COMMA floatnum shorten_range
	{
		assign_float(MAXPROSHORT, $2, Currstruct_p);
	};

shorten_range:	/* empty */
	{
	}

	|
	T_COMMA value T_COMMA value
	{
		if ($2 > $4) {
			l_warning(Curr_filename, yylineno,
				"distance to begin shortening protruding stems cannot be greater than ending distance; making them equal");
			$4 = $2;
		}
		assign_int(BEGPROSHORT, $2, Currstruct_p);
		assign_int(ENDPROSHORT, $4, Currstruct_p);
	};

pagesize:
	T_PAGESIZE

	|
	pitch num
	{
		if ($1 == (int) 'a' && $2 >= 4 && $2 <= 6) {
			switch ($2) {
			case 4:
				$$ = PS_A4;
				break;
			case 5:
				$$ = PS_A5;
				break;
			case 6:
				$$ = PS_A6;
				break;
			default:
				pfatal("impossible pagesize");
				break;
			}
		}
		else {
			yyerror("unsupported pagesize");
		}
	}
	;

orientation:
	/* empty */
	{
		$$ = 0;
	}

	|
	T_ORIENTATION
	{
		$$ = ($1 == O_PORTRAIT ? 0 : 1);
	}
	;

vcombine_value:
	/* empty */
	{
		assign_vcombine(VC_NOOVERLAP, Currstruct_p);
	}

	|
	voice_spec opt_vcombine_qualifier
	{
		assign_vcombine($2, Currstruct_p);
		Defining_multiple = NO;
	}
	;

opt_vcombine_qualifier:
	/* empty */
	{
		$$ = VC_NOOVERLAP;
	}

	|
	T_VCOMBVAL
	{
		$$ = $1;
	}
	;

end_type:
	T_ENDTYPE
	{
		$$ = $1;
	}

	|
	T_BLOCKHEAD
	{
		if ($1 != C_TOP) {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
		$$ = ENDING_TOP;
	};

updown:	T_UP
	{
		$$ = 1;
	}

	|
	T_DOWN
	{
		$$ = -1;
	};

majmin:
	{
		$$ = NO;
	}

	|
	T_XPOS_INT
	{
		switch ($1) {
		case MINOR:
			$$ = YES;
			break;
		case MAJOR:
			$$ = NO;
			break;
		default:
			$$ = NO;
			yyerror("must be major or minor");
			break;
		}
	};

pedstyle:	T_LINE
	{
		$$ = P_LINE;
	}

	|
	T_ALT T_PEDSTAR
	{
		$$ = P_ALTPEDSTAR;
	}

	|
	T_PEDSTAR
	{
		$$ = P_PEDSTAR;
	};

transpose_interval: T_XPOS_INT
	;

title:	T_TITLE tfont title_size titleA opt_semi T_NEWLINE
	{
	};

tfont:	font_family title_font
	{
		if ($2 == FONT_UNKNOWN) {
			$2 = Curr_font;
		}
		if ($1 == FAMILY_DFLT) {
			/* user didn't specify a family. Use the current
			 * family, unless we don't have one yet, in which
			 * case look it up */
			if (Curr_family == FAMILY_DFLT) {
				Curr_family = Score.fontfamily;
			}
			Titlefont =  Curr_family + $2;
		}
		else {
			Titlefont = $1 + $2;
		}
	};

title_font:
	{
		$$ = FONT_UNKNOWN;
	}

	|
	T_FONT
	{
		$$ = $1;
	};

font_family:
	{
		$$ = FAMILY_DFLT;
	}

	|
	T_FFAMILY {
		$$ = $1;
	};

title_size:
	{
		Titlesize = Curr_size;
	}

	|
	T_LPAREN num T_RPAREN
	{
		Titlesize = $2;
	};

titleA:
	string 
	{
		Extra = extra_needed(Titlefont, Titlesize, $1);
		proc_printcmd(J_CENTER, (struct INPCOORD *) 0, $1,
				Titlefont, Titlesize, YES, NO, Extra);
	}

	|
	string string
	{
		double tmp_extra;

		/* if one is taller than the default for the font/size,
		 * figure out how much more to add on */
		/* can't use MAX passing extra_needed() as arguments because
		 * it would get executed multiple times, which is a problem */
		Extra = extra_needed(Titlefont, Titlesize, $1);
		tmp_extra = extra_needed(Titlefont, Titlesize, $2);
		Extra = MAX(Extra, tmp_extra);


		proc_printcmd(J_LEFT, (struct INPCOORD *) 0, $1,
				Titlefont, Titlesize, YES, NO, Extra);
		proc_printcmd(J_RIGHT, (struct INPCOORD *) 0, $2,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
	}

	|
	string string string
	{
		double tmp_extra;

		Extra = extra_needed(Titlefont, Titlesize, $1);
		tmp_extra = extra_needed(Titlefont, Titlesize, $2);
		Extra = MAX(Extra, tmp_extra);
		tmp_extra = extra_needed(Titlefont, Titlesize, $3);
		Extra = MAX(Extra, tmp_extra);

		proc_printcmd(J_LEFT, (struct INPCOORD *) 0, $1,
				Titlefont, Titlesize, YES, NO, Extra);
		proc_printcmd(J_CENTER, (struct INPCOORD *) 0, $2,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
		proc_printcmd(J_RIGHT, (struct INPCOORD *) 0, $3,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
	};

paragraph:	opt_paratype T_PARAGRAPH tfont title_size string opt_semi T_NEWLINE
	{
		char *p;		/* pointer through the string */
		int backslash_count;	/* how many backslashes in a row */
		int font, size;		/* at end of \n-delimited segment */
		char *string_start;	/* where current segment begins */

		/* Convert unescaped input newlines to spaces.
		 * The +2 in the 'for' initialization is to skip font/size.
		 */
		backslash_count = 0;
		string_start = $5;
		/* If we are running on a file using \r or \r\n for
		 * line separator, normalize to \n instead */
		for (p = string_start + 2; *p != 0; p++) {
			if (*p == '\r') {
				if ( *(p+1) != '\n') {
					/* \r by itself. Use \n instead */
					*p = '\n';
				}
				else {
					/* delete the \r */
					char *src, *dest;
					for (src = p + 1, dest = p;  ; src++, dest++) { 
						*dest = *src;
						if (*src == '\0') {
							break;
						}
					}
				}
			}
		}

		font = Titlefont;
		size = Titlesize;
		for (p = string_start + 2; *p != 0; p++) {
			/* Real newlines not preceeded by an odd number
			 * of backslashes are converted to spaces. */
			if (*p == '\n' && (backslash_count & 0x1) == 0) {
				*p = ' ';
			}

			/* Count up number of consecutive backslashes.
			 * Also, if user explicitly uses \n, split into
			 * a new paragraph
			 */
			if (*p == '\\') {
				backslash_count++;
				if (((backslash_count & 0x1) == 1)
						&& ( *(p+1) == 'n') ) {
					/* Need to split.
					 * Terminate the current string
					 * where we are in it, and process it,
					 * then arrange to continue processing
					 * on the rest of the string.
					 */
					*p = '\0';
					(void) fix_string(string_start,
						font, size,
						Curr_filename, yylineno);
					proc_printcmd($1, (struct INPCOORD*) 0,
						string_start, font, size,
						YES, NO, 0.0);

					/* The user could change font/size
					 * in mid-string with \f() and/or \s().
					 * so we have to determine what they
					 * are at the end of the current segment
					 * and use that for making the copy
					 * of the remainder of the string.
					 * The original copy of the remainder
					 * of the string gets "memory leaked,"
					 * but this is a rare case, so not
					 * worth worrying about.
					 */
					end_fontsize(string_start, &font, &size);
					string_start = copy_string(p+2, font, size);
					p = string_start + 2;
				}
			}
			else {
				backslash_count = 0;
			}
		}
		(void) fix_string(string_start, font, size,
					Curr_filename, yylineno);
		proc_printcmd($1, (struct INPCOORD*) 0, string_start,
					font, size, YES, NO, 0.0);
	};

opt_paratype:
	/* empty */
	{
		/* Use same paragraph time as last time (or default if
		 * this is the first paragraph ever in this song.) */
		$$ = Curr_paratype;
	}

	|
	T_PARATYPE
	{
		Curr_paratype = $1;
		$$ = $1;
	};

paramname:
	T_NUMVAR

	|
	T_YESNOVAR

	|
	T_SCORESEP

	|
	T_STAFFLINES

	|
	T_TIMEUNIT

	|
	T_SWINGUNIT

	|
	T_PAD

	|
	T_BARSTLISTVAR

	|
	T_FNUMVAR

	|
	T_2FNUMVAR

	|
	T_SHORTEN

	|
	T_UNITS

	|
	T_ENDSTYLE

	|
	T_REH_STYLE

	|
	T_PEDSTYLE

	|
	T_VVAR

	|
	T_VCOMBINE

	|
	T_SCOREPAD

	|
	T_RANGELISTVAR

	|
	T_ORDER

	|
	T_USEACCS

	|
	T_KEY

	|
	T_TIME

	|
	T_TRANSPOSE

	|
	T_RATNUMLISTVAR

	|
	T_VISVAR

	|
	T_STRVAR

	|
	T_FONTVAR

	|
	T_FAMILY

	|
	T_CLEFVAR
	;

string:	T_STRING
	{
		/* strip the quotes from the string and make a copy for later use */
		if (Curr_family == FAMILY_DFLT) {
			Curr_family = Score.fontfamily;
		}
		$$ = copy_string(stripquotes(yytext), Curr_family + Curr_font,
								Curr_size);
	}

	|
	string T_PLUS T_STRING
	{
		char *old_string;

		/* append new string part to existing part */
		old_string = $1;
		/* new string part has quotes, so subtract 2 from needed length,
		 * but need space for null, so overall need 1 less. */
		MALLOCA(char, $$, strlen(old_string) + strlen(yytext) - 1);
		sprintf($$, "%s%s", old_string, stripquotes(yytext));
		FREE(old_string);
	};

music_input:	noteinfo T_NEWLINE
	{
		(void) contextcheck(C_MUSIC, "music information");
		tssv_line();
		/* reset flags for next voice */
		Defining_multiple = NO;
		Curr_grpsyl_p = (struct GRPSYL *) 0;
		Good_till_canceled = NO;
		free_extra_time();
		Prev_grpsyl_p = 0;
	};

noteinfo:	init_staff sv_spec notecolon groupinfo
	{
		/* Here's the deal: the staff/voice specification
		 * could be lists of staff ranges and/or voices.
		 * We need a separate copy of the groupinfo
		 * for each staff/voice. So as we parse the sv_spec,
		 * we remember which staffs and voices are specified,
		 * and keep a pointer to the first STAFF in the measure.
		 * Then after collecting the groupinfo,
		 * we make copies of the group info for each staff/voice given.
		 * It has to be copies rather than just a pointer to the
		 * same stuff, because the coordinates will be different
		 * for each staff/voice. */

		/* attach the groupinfo to each of the STAFF structs */
		link_groups();
	}

	|
	T_MULTIREST num opt_semi
	{
		add_multirest($2);
		/* If there are alternating time signatures, figure out
		 * which one should be in effect at the end of the multirest */
		if (Alt_timesig_list != 0) {
			int m;

			/* The Next_alt_timesig is already pointing to
			 * the value for after the first measure in the
			 * multirest, so have to subtract 1 here */
			for (m = 0; m < $2 - 1; m++) {
				if (Next_alt_timesig == 0) {
					/* Shouldn't really ever get here,
					 * since other code should do the
					 * wrap-around, but just in case... */
					Next_alt_timesig = Alt_timesig_list;
					continue;
				}
				do {
					Next_alt_timesig++;
					if (*Next_alt_timesig == TSR_END) {
						Next_alt_timesig = Alt_timesig_list;
						break;
					}
				} while (*Next_alt_timesig != TSR_ALTERNATING);
			}
			if (*Next_alt_timesig == TSR_ALTERNATING) {
				Next_alt_timesig++;
			}
		}
		Currstruct_p = (struct MAINLL *) 0;
	}

	|
	lyricinfo
	{
		free_staffrange();
	};

init_staff:
	{
		begin_range(PL_UNKNOWN);
	};

sv_spec:	voice_id
	{
		Chord_at_a_time = NO;
	}

	|
	sv_map_list
	{
		Chord_at_a_time = YES;
	}
	;

sv_map_list:	sv_map_entry

	|
	sv_map_list sv_map_entry
	;

sv_map_entry:	map_start sv_map_item_list map_finish
	;

map_start:	T_LBRACKET
	{
		begin_map();
	};

map_finish:	T_RBRACKET
	{
		end_map();
	};

sv_map_item_list:	voice_id
	{
		map_item();
	}

	|
	sv_map_item_list T_SEMICOLON voice_id
	{
		map_item();
		Defining_multiple = YES;
	};

voice_id:	sv_item
	{
		/* initialize grpsyl info */
		Curr_grpsyl_p = Last_grpsyl_p = (struct GRPSYL *) 0;
	}

	|
	voice_id sv_and sv_item
	;

sv_and:	T_AMPERSAND
	{
		/* prepare for a new set of ranges */
		begin_range(Place);
		Defining_multiple = YES;
	};

sv_item:	grp_stafflist
	{
		/* data is for staff with only one voice */
		save_vno_range(1, 1);
		add_to_sv_list();
	}

	|
	grp_stafflist voice_spec
	{
		if ($2 == YES) {
			add_to_sv_list();
		}
	}
	;

notecolon:
	T_COLON
	{
		/* set flag as to whether we are processing tab staff
		 * information or normal staff info */
		Doing_tab_staff = is_tab_range();
	};

grp_stafflist:	grp_staff_item

	|
	grp_stafflist T_COMMA grp_staff_item
	{
		/* several staffs have the same note data */
		Defining_multiple = YES;
	};

grp_staff_item:	staffnum grp_staff_range
	{
		/* remember which staffs we are currently getting data for */
		/* if only a single number, not a range, staff_range will be
		 * 0, so use staffnum as high and low of range */
		save_staff_range($1, ($2 == 0 ? $1 : $2));
	};

grp_staff_range:
	{
		/* empty - no range, just a single number */
		$$ = 0;
	}

	|
	T_DASH staffnum
	{
		Defining_multiple = YES;
		$$ = $2;
		if ($2 == 0) {
			yyerror("staff of 0 is illegal");
		}
	};

voice_spec:	
	voicenum voice_tail
	{
		/* note data applies to a single voice */
		if (rangecheck($1, MINVOICES, MAXVOICES, "voice") == YES) {
			save_vno_range($1, $1);
			$$ = YES;
		}
		else {
			$$ = NO;
		}
	}

	|
	voicenum T_DASH voicenum voice_tail
	{
		/* note data applies to range of voices */
		if (rangecheck($1, MINVOICES, MAXVOICES, "voice") == YES &&
				rangecheck($3, MINVOICES, MAXVOICES, "voice") == YES) {
			if ($3 < $1) {
				l_yyerror(Curr_filename, yylineno,
					"end of voice range is smaller than beginning");
				$$ = NO;
			}
			else {
				save_vno_range($1, $3);
				Defining_multiple = YES;
				$$ = YES;
			}
		}
		else {
			$$ = NO;
		}
	};

voice_tail:
	/* empty */

	|
	T_COMMA voice_spec
	{
		Defining_multiple = YES;
	};

groupinfo:	groupdata

	|
	groupinfo groupdata
	;

groupdata:
	ntuplet

	|
	mid_meas_param grp_attributes group other_attr T_SEMICOLON
	{
		/* Associate any TIMEDSSVs with this group */
		if (Curr_tssv_p != 0) {
			tssv_setgrpsyl(Curr_grpsyl_p);
			Curr_tssv_p = 0;
		}

		/* if no custom beaming indicated, fill in based on previous. */
		if ( Curr_grpsyl_p->beamloc == NOITEM) {
			setbeamloc(Curr_grpsyl_p, Last_grpsyl_p);
		}

		/* later we check that if there is an 'm', it is
		 * the only thing in the measure */

		link_notegroup(Curr_grpsyl_p, Last_grpsyl_p);
		Curr_marklist = 0;
		/* Save pointer to this group, in case there are additive
		 * times, and the next group gets its time value from this
		 * group. In that case, we need this group, not the last one
		 * of the added groups. */
		Prev_grpsyl_p = Curr_grpsyl_p;
		/* Add groups if there were additive time values. */
		if (Extra_time_p != 0) {
			Curr_grpsyl_p = expandgrp(Curr_grpsyl_p, Extra_time_p);
			/* If there was a custom beam ebm on something with
			 * additive time, we need make sure ENDITEM is on only
			 * the last of the expanded groups. */
			if (Curr_grpsyl_p != Prev_grpsyl_p
					&& Prev_grpsyl_p->beamloc == ENDITEM) {
				struct GRPSYL *g_p;
				for (g_p = Prev_grpsyl_p; g_p != Curr_grpsyl_p;
							g_p = g_p->next) {
					g_p->beamloc = INITEM;
				}
				Curr_grpsyl_p->beamloc = ENDITEM;
			}
			Last_grpsyl_p = Curr_grpsyl_p->prev;
		}
		Curr_timelist_p = 0;
	};

mid_meas_param:
	/* empty */

	|
	mid_meas_list
	;

mid_meas_list:
	mid_meas_item

	|
	mid_meas_list mid_meas_item
	;

mid_meas_item:
	T_L_DBLANGLE mm_context mm_params T_R_DBLANGLE
	{
		if (Chord_at_a_time == YES) {
			yyerror("mid-measure changes not allowed on chord-at-a-time input");
		}
	}
	;

mm_context:
	T_SCORE
	{
		Curr_tssv_p = tssv_create(C_SCORE);
	}

	|
	T_STAFF
	{
		Curr_tssv_p = tssv_create(C_STAFF);
	}

	|
	T_VOICE
	{
		Curr_tssv_p = tssv_create(C_VOICE);
	}
	;

mm_params:
	mm_param_item

	|
	mm_params T_SEMICOLON mm_param_item
	;

mm_param_item:
	T_CLEFVAR T_EQUAL T_CLEF
	{
		if (Curr_grpsyl_p != 0 && Curr_grpsyl_p->grpsyl == GS_GROUP
				&& Curr_grpsyl_p->grpvalue == GV_ZERO) {
			yyerror("mid-measure clef change not allowed after grace note\n");
		}
		tssv_update(Curr_tssv_p, $1, $3);
	}

	|
	T_NUMVAR T_EQUAL num
	{
		switch ($1) {

		case DEFOCT:
		case RELEASE:
			tssv_update(Curr_tssv_p, $1, $3);
			break;

		default:
			l_warning(Curr_filename, yylineno,
				"parameter type cannot be changed mid-measure; ignoring");
			break;
		}
	}
	;

grp_attributes:	styleinfo time_val
	;

styleinfo:
	alloc_grpsyl
	{
		if (Good_till_canceled == NO) {
			/* use default attributes */
			Curr_grpsyl_p->grpvalue = GV_NORMAL;
			Curr_grpsyl_p->grpsize = GS_NORMAL;
			Curr_grpsyl_p->headshape = HS_UNKNOWN;
		}
		else {
			/* re-use previous style */
			copy_attributes(Curr_grpsyl_p, Last_grpsyl_p);
		}
	}

	|
	T_LBRACKET T_DASH T_RBRACKET alloc_grpsyl
	{
		Good_till_canceled = NO;
	}

	|
	T_LBRACKET alloc_grpsyl marklist T_RBRACKET gtc extra_styleinfo
	{
	};

gtc:
	{
		Good_till_canceled = NO;
	}

	|
	T_DOT T_DOT T_DOT
	{
		Good_till_canceled = YES;
	};

extra_styleinfo:
	{
	}

	|
	extra_styleinfo T_LBRACKET marklist T_RBRACKET
	{
		if (Good_till_canceled == YES) {
			yyerror("can't use ... and more than one [] on same chord");
		}
	};

alloc_grpsyl:
	{
		/* NULL token to cause allocation of a GRPSYL struct */
		Last_grpsyl_p = Curr_grpsyl_p;
		Curr_grpsyl_p = newGRPSYL(GS_GROUP);
	};

marklist:
	{
		/* an empty marklist means we should use
		 * the same attributes as the last time */
		copy_attributes(Curr_grpsyl_p, Last_grpsyl_p);
	}

	|
	marklistA
	;

marklistA: markitem

	|
	marklist T_SEMICOLON markitem
	;

markitem:	T_WITH alloc_marklist marks 
	{
		Curr_grpsyl_p->nwith = (short) Item_count;
		Curr_grpsyl_p->withlist = Curr_marklist;
	}

	|
	grp_locsave

	|
	notetype

	|
	padding

	|
	dir
	{
		if (Curr_grpsyl_p->stemdir != UNKNOWN &&
					Curr_grpsyl_p->stemdir != $1) {
			l_warning(Curr_filename, yylineno,
				"both stem directions specified; using last instance");
		}
		if (Doing_tab_staff == YES) {
			l_warning(Curr_filename, yylineno,
				"stem direction specification is pointless on tab staff; ignoring");
		}
		else {
			Curr_grpsyl_p->stemdir = $1;
		}
	}

	|
	T_FNUMVAR opt_minus num
	{
		if ($1 != DIST) {
			l_warning(Curr_filename, yylineno,
				"unexpected parameter name; ignoring");
		}
		else {
			if (Curr_grpsyl_p->restdist != NORESTDIST &&
					Curr_grpsyl_p->restdist != $2 * $3) {
				l_warning(Curr_filename, yylineno,
						"more than one dist value specified; using last instance");
			}
			Curr_grpsyl_p->restdist = $2 * $3;
		}
	}

	|
	T_HO hor_offset
	{
		if (Curr_grpsyl_p->ho_usage != HO_NONE &&
				Curr_grpsyl_p->ho_usage != $2) {
			l_warning(Curr_filename, yylineno,
				"More than one ho type specified; last instance used");
		}
		Curr_grpsyl_p->ho_usage = $2;
	}

	|
	T_HS string
	{
		if (Curr_grpsyl_p != 0) {
			if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
				l_warning(Curr_filename, yylineno,
					"multiple head shapes specified, using last");
			}
			/* +2 to skip past extraneous font/size */
			if (Doing_tab_staff == YES && strcmp($2 + 2, "allx") != 0) {
				yyerror("allx is the only headshape allowed on tab staffs");
			}
			if ((Curr_grpsyl_p->headshape = get_shape_num($2 + 2))
					== HS_UNKNOWN) {
				l_yyerror(Curr_filename, yylineno,
					"'%s' is not a valid head shape name",
					ascii_str($2, YES, NO, TM_NONE));
			}
		}
		else {
			pfatal("Curr_grpsyl_p was null for setting hs");
		}
		/* We don't need the name anymore; we have its corresponding
		 * number that we use internally. */
		FREE($2);
	}

	|
	T_LEN floatnum
	{
		if (Curr_grpsyl_p->stemlen != STEMLEN_UNKNOWN &&
				fabs(Curr_grpsyl_p->stemlen - $2 * STEPSIZE) > 0.0001) {
			l_warning(Curr_filename, yylineno,
				"more than one len specified; using last instance");
		}
		Curr_grpsyl_p->stemlen = $2 * STEPSIZE;
	}

	|
	T_SLASHMARK num
	{
		if ($2 <= 0) {
			yyerror("slash number must be > 0");
		}
		else if ($2 > 8) {
			/* we decided 8 slashes would be at least 256th note
			 * or shorter, and 256th is the shortest note we
			 * support, 8 slashes is plenty */
			l_warning(Curr_filename, yylineno,
				"number of slashes only allowed to be <= 8; limiting to 8");
			$2 = 8;
		}
		if (Curr_grpsyl_p->slash_alt > 0 &&
					Curr_grpsyl_p->slash_alt != $2) {
			l_warning(Curr_filename, yylineno,
				"more than one slash value specified; using last instance");
		}
		if (Curr_grpsyl_p->slash_alt < 0) {
			yyerror("only one slash/alt allowed per group");
		}
		Curr_grpsyl_p->slash_alt = $2;
	};

hor_offset:
	T_PLUS
	{
		$$ = HO_RIGHT;
	}

	|
	T_DASH
	{
		$$ = HO_LEFT;
	}

	|
	opt_plus_or_minus floatnum
	{
		/* We want to silently accept identical values specified
		 * more than once, so use fuzzy compare to deal with roundoff */
		if (Curr_grpsyl_p->ho_usage == HO_VALUE &&
				fabs(Curr_grpsyl_p->ho_value - $1 * $2) > 0.0001) {
			l_warning(Curr_filename, yylineno,
				"More than one ho value specified; last instance used");
		}
		Curr_grpsyl_p->ho_value = $1 * $2;
		$$ = HO_VALUE;
	};

opt_plus_or_minus:
	/* empty */
	{
		$$ = 1;
	}

	|
	plus_or_minus
	{
		$$ = $1;
	};

other_attr:

	|
	attr_list
	;

attr_list:
	attr_item

	|
	attr_list T_COMMA attr_item
	;

attr_item:
	beammark

	|
	T_SLOPE opt_minus floatnum
	{
		$3 *= $2;
		if (fabs(Curr_grpsyl_p->beamslope - NOBEAMANGLE) > 0.01) {
			l_warning(Curr_filename, yylineno,
				"multiple slope values specified; using last one");
		}
		if (frangecheck($3, MINBEAMANGLE, MAXBEAMANGLE, "slope") == YES) {
			Curr_grpsyl_p->beamslope = $3;
		}
	}

	|
	T_ALT num
	{
		if ($2 <= 0) {
			yyerror("alt number must be > 0");
		}
		if (Curr_grpsyl_p->slash_alt != 0) {
			yyerror("only one slash/alt allowed per group");
		}
		if (Doing_tab_staff == YES) {
			l_warning(Curr_filename, yylineno,
				"alt not allowed on tablature staff; ignoring");
		}
		else {
			/* Should at least keep small enough
			 * that 1 shifted left by this still fits
			 * in a 16-bit int. For a while we limited to 15,
			 * but for slashes, we limit to 8, and
			 * that seems like more than enough for here too. */
			if ($2 > 8) {
				l_warning(Curr_filename, yylineno,
					"alt value too large; limiting to 8");
				$2 = 8;
			}
			Curr_grpsyl_p->slash_alt = -$2;
		}
	}

	|
	T_SLUR opt_dir
	{
		int n;

		/* add a slur to each note in the chord. Don't know
		 * which pitch to slur to yet, just that it will be to
		 * the "matching" note in the next chord, so use
		 * special magic pitch of 'M' to convey this */
		for (n = 0; n < Curr_grpsyl_p->nnotes; n++) {
			add_slurto(Curr_grpsyl_p, 'M', USE_DFLT_OCTAVE, n, $1);
			set_slurdir(Curr_grpsyl_p, -1, $2);
		}
	}

	|
	T_TIE opt_dir
	{
		int n;

		Curr_grpsyl_p->tie = YES;
		for (n = 0; n < Curr_grpsyl_p->nnotes; n++) {
			Curr_grpsyl_p->notelist[n].tie = YES;
			Curr_grpsyl_p->notelist[n].tiestyle = $1;
			Curr_grpsyl_p->notelist[n].tiedir = $2;
		}
	};

alloc_marklist:
	{
		if (Curr_marklist == 0) {
			/* allocate space for a list of marks */
			MALLOCA(char *, Curr_marklist, ITEMS);
			Item_count = 0;
			Max_items = ITEMS;
		}
	};

opt_dir:
	{
		$$ = UNKNOWN;
	}

	|
	dir
	{
		$$ = $1;
	}
	;

marks:	mark

	|
	marks T_COMMA mark
	;

mark:	mark_item
	{
		/* if too many items, get some more space */
		if (Item_count >= Max_items) {
			Max_items += ITEMS;
			if ((Curr_marklist = (char **) realloc(Curr_marklist,
						Max_items * sizeof(char **)))
						== (char **) 0) {
				l_no_mem(__FILE__, __LINE__);
			}
		}
		Curr_marklist[Item_count++] = $1;
	};

mark_item:	shorthand_mark
	{
		MALLOCA(char, $$, strlen($1) + 3);
		$$[0] = FONT_TR;
		$$[1] = DFLT_SIZE;
		sprintf($$ + 2, $1);
	}

	|
	string
	;

shorthand_mark:
	T_DOT
	{
		$$ = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(dot)" : "\\(smdot)");
	}

	|
	T_DASH
	{
		$$ = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(leg)" : "\\(smleg)");
	}

	|
	T_R_ANGLE
	{
		$$ = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(acc_gt)" : "\\(smacc_gt)");
	}

	|
	T_HAT
	{
		$$ = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(acc_hat)" : "\\(smacc_hat)");
	};

notetype:	T_CUE
	{
		if (Curr_grpsyl_p->grpvalue == GV_ZERO) {
			l_warning(Curr_filename, yylineno,
				"both grace and cue specified; using last instance");
		}
		Curr_grpsyl_p->grpvalue = GV_NORMAL;
		Curr_grpsyl_p->grpsize = GS_SMALL;
	}

	|
	T_GRACE
	{
		if (Curr_grpsyl_p->grpsize == GS_SMALL
					&& Curr_grpsyl_p->grpvalue != GV_ZERO) {
			l_warning(Curr_filename, yylineno,
				"both cue and grace specified; using last instance");
		}

		Curr_grpsyl_p->grpsize = GS_SMALL;
		Curr_grpsyl_p->grpvalue = GV_ZERO;
	}

	|
	T_DIAM
	{
		if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
			l_warning(Curr_filename, yylineno,
				"diam is overriding previous headshapes specification");
		}
		Curr_grpsyl_p->headshape = get_shape_num("diam");
	}

	|
	T_XNOTE
	{
		if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
			l_warning(Curr_filename, yylineno,
				"xnote is overriding previous headshapes specification");
		}
		Curr_grpsyl_p->headshape = get_shape_num(
					Doing_tab_staff == YES ? "allx" : "x");
	};

dir:
	T_UP
	{
		$$ = UP;
	}

	|
	T_DOWN
	{
		$$ = DOWN;
	};

beammark:
	T_BM crossbeam
	{
		/* custom beaming cannot be nested */
		struct GRPSYL *prev_like_gs_p;
		/* Find previous group with same grace-ness. Can't use
		 * prevsimilar() here, since it looks at grpcont too. */
		for (prev_like_gs_p = Last_grpsyl_p; prev_like_gs_p != 0 &&
			prev_like_gs_p->grpvalue != Curr_grpsyl_p->grpvalue;
			prev_like_gs_p = prev_like_gs_p->prev) {
			;
		}
		if (prev_like_gs_p != 0 &&
				((prev_like_gs_p->beamloc == STARTITEM)
				|| (prev_like_gs_p->beamloc == INITEM)) ) {
			yyerror("custom beaming may not be nested");
		}
		/* Non-custom beaming, if any, is done later, in do_bar() */
		else {
			/* begin custom beaming */
			Curr_grpsyl_p->beamloc = STARTITEM;
			if (Curr_grpsyl_p->basictime < 8 && $2 == CS_SAME) {
				yyerror("beamed notes must be 8th or shorter");
			}
		}
		Curr_grpsyl_p->beamto = $2;
	}

	|
	T_ESBM
	{
		Curr_grpsyl_p->breakbeam = YES;
	}

	|
	T_EBM
	{
		struct GRPSYL *last_nongrace_p;

		/* find the previous group, skipping grace groups */
		for (last_nongrace_p = Last_grpsyl_p;
				last_nongrace_p != (struct GRPSYL *) 0
				&& last_nongrace_p->grpvalue == GV_ZERO;
				last_nongrace_p = last_nongrace_p->prev) {
			;
		}

		/* check that a custom beam is in progress */
		if ((last_nongrace_p == (struct GRPSYL *) 0)
				|| ((last_nongrace_p != (struct GRPSYL *) 0)
				&& (last_nongrace_p->beamloc != STARTITEM)
				&& (last_nongrace_p->beamloc != INITEM) )) {
			l_warning(Curr_filename, yylineno,
				"'ebm' unexpected: no custom beaming in progress; ignoring");
		}
		else {
			Curr_grpsyl_p->beamloc = ENDITEM;
			Curr_grpsyl_p->beamto = last_nongrace_p->beamto;
			if (Curr_grpsyl_p->basictime < 8
					&& Last_grpsyl_p->beamto == CS_SAME) {
				yyerror("beamed notes must be 8th or shorter");
			}
		}
	};

time_val:
	{
		/* Use the same as last time or default timeunit.
		 * If that involves additive time values,
		 * save pointer to that info in Extra_time_p for later use. */
		Extra_time_p = copy_timeunit(Curr_grpsyl_p, Prev_grpsyl_p,
						Extra_time_p);
	}

	|
	time_item
	{
		if (Curr_grpsyl_p->is_meas == YES) {
			Curr_grpsyl_p->fulltime = Score.time;
		}
		else {
			Curr_grpsyl_p->fulltime = $1;
		}
	};

crossbeam:
	{
		$$ = CS_SAME;
	}

	|
	T_WITH T_STAFF T_PLACE
	{
		CSBused = YES;
		switch ($3) {
		case PL_ABOVE:
			$$ = CS_ABOVE;
			break;
		case PL_BELOW:
			$$ = CS_BELOW;
			break;
		default:
			yyerror("bm with staff must be 'above' or 'below'");
			$$ = CS_SAME;
		}
	};

basic_time_val:	T_MULTIWHOLE
	{
		$$.n = $1;
		$$.d = 1;

		/* if filling in a GRPSYL struct, need to fill in basic time,
		 * could also be here due to beamstyle, in which case the
		 * Curr_grpsyl_p will be NULL, or when getting tuplet duration,
		 * in which case flag will be set */
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			/* If we are gathering basictime as part of a list
			 * of additive times, we will save the value a few
			 * lines down from here. But in the normal case,
			 * we set basictime in the current GRPSYL. */
			if (Extra_time_p == 0) {
				/* 1/2 is 0 internally. 1/4 is -1 internally */
				 Curr_grpsyl_p->basictime = ($1 == 2 ? 0 : -1);
			}
		}
		/* If doing additive times, need to save value. */
		if (Extra_time_p != 0) {
			Extra_basictime = ($1 == 2 ? 0 : -1);
		}
	}

	|
	num
	{
		/* check that a power of two from 1 to MAXBASICTIME */
		if (power_of2check($1, "note basic time value") == NO) {
			/* force to a power to two, so that other code
			 * (like in expandgrp) that expect a sane value
			 * to not blow up. This may lead to a somewhat
			 * misleading "time does not add up to time					 * signature message, but we don't know what
			 * time they really meant, and this is better
			 * than pfataling. */
			$1 = 2;
		}
		/* can't use rangecheck here because the error message would
		 * say 0 and -1 are valid times, which is only true internally--
		 * the user has to use 1/2, 1/4 or m. */
		if ($1 < MINBASICTIME || $1 > MAXBASICTIME) {
			l_yyerror(Curr_filename, yylineno,
				"time value must be between 1 and %d, or 1/2 or 1/4 or m",
				MAXBASICTIME);
			/* If user entered outrageous value, clamp to
			 * something semi-reasonable, else we could get
			 * rational overflow, and pfatal, even though it
			 * it is really a user error. */
			if ($1 > MAXBASICTIME * 4) {
				$1 = MAXBASICTIME * 4;
			}
		}
		$$.n = 1;
		/* avoid division by zero */
		if ($1 == 0) {
			$1 = 1;
		}
		$$.d = $1;
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			if (Curr_timelist_p == 0) {
				Curr_grpsyl_p->basictime = $1;
			}
		}
		/* If doing additive times, need to save value */
		if (Extra_time_p != 0) {
			Extra_basictime = $1;
		}
	};

opt_more_time:
	/* empty */
	{
		free_extra_time();
	}

	|
	opt_more_time more_time_item 
	;

more_time_item:
	plus_or_minus alloc_timelist basic_time_val dots
	{
		/* Set basictime to what we saved in basic_time_val rule,
		 * then calculate fulltime from that and number of dots. */
		Curr_timelist_p->basictime = Extra_basictime;
		Curr_timelist_p->fulltime = calcfulltime($3, $4);
		/* handle subtracted times by negating the fulltime */
		if ($1 == -1) {
			Curr_timelist_p->fulltime = rneg(Curr_timelist_p->fulltime);
		}
	}
	;

alloc_timelist:
	/* empty -- just for allocating TIMELIST struct */
	{
		struct TIMELIST *timelist_p;

		MALLOC(TIMELIST, timelist_p, 1);
		/* Add to end of linked list */
		timelist_p->next = 0;
		/* Init fulltime to something to avoid garbage if there
		 * in a user input error */
		timelist_p->fulltime = Zero;
		if (Extra_time_p == 0) {
			Last_alloced_timelist_p = Extra_time_p = timelist_p;
		}
		else {
			Curr_timelist_p->next = timelist_p;
		}
		/* Keep track of where to append next item to list, if any */
		Curr_timelist_p = timelist_p;
	}
	;

opt_m:	
	{
		User_meas_time = NO;
	}

	|
	T_LET_M
	{
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			Curr_grpsyl_p->is_meas = YES;
			User_meas_time = YES;
		}
		else {
			yyerror("'m' is not valid here");
		}
	};

dots:
	{
		$$ = 0;
	}

	|
	dots T_DOT
	{
		/* count up the number of dots */
		$$ = $1 + 1;
	};

padding: padspec
	{
		/* We can't distinguish between the default 0.0 and
		 * if user explicitly sets to 0.0, so if they set to 0.0
		 * then to something else, we won't catch that as setting
		 * twice, but that shouldn't be common. Besides, if they
		 * expect them to be additive, adding to zero will work
		 * as they expect... Because of roundoff, we can't do
		 * exact compare for specifying the same value more than once,
		 * so treat as identical if pretty close. */
		if (Curr_grpsyl_p->padding != 0.0 && 
				fabs(Curr_grpsyl_p->padding - ($1 * STEPSIZE))
				> 0.0001) {
			l_warning(Curr_filename, yylineno,
				"padding specified more than once; using last instance");
		}
		Curr_grpsyl_p->padding = $1 * STEPSIZE;
	}
	;

padspec: T_PAD opt_minus floatnum
	{
		$$ = (float) $2 * $3;
	};

opt_minus:
	{
		/* no sign--must be a positive number */
		$$ = 1;
	}

	|
	T_DASH
	{
		/* user wants a negative number */
		$$ = -1;
	};

group:	notelist
	{
		Curr_grpsyl_p->grpcont = GC_NOTES;
		if ($1 == 0) {
			/* no notes listed, use same as previous group */
			copy_notes(Curr_grpsyl_p, Last_grpsyl_p);
		}
		else {
			resize_notelist(Curr_grpsyl_p);
		}
	}

	|
	special_group

	;

special_group:
	extra_attr note_opts
	;

extra_attr:
	spec_grp_alloc extra_item
	;

spec_grp_alloc:
	{
		/* allocate GRPSYL for the case where everything is defaulted
		 * from the previous group, with just an extra attribute
		 * like ? or ~ specified. */
		if (Curr_grpsyl_p == (struct GRPSYL *) 0) {
			/* shouldn't ever happen, but just in case... */
			Curr_grpsyl_p = newGRPSYL(GS_GROUP);
		}
		if (Last_grpsyl_p != (struct GRPSYL *) 0 &&
					Last_grpsyl_p->nnotes >= 1) {
			copy_notes(Curr_grpsyl_p, Last_grpsyl_p);
		}
		else if (Last_grpsyl_p != (struct GRPSYL *) 0 && 
				Last_grpsyl_p->grpcont == GC_REST) {
			Curr_grpsyl_p->grpcont = GC_REST;
		}
	};

extra_item:
	T_QUESTION
	{
		if (Curr_grpsyl_p->grpcont == GC_NOTES &&
					Curr_grpsyl_p->nnotes > 0) {
			Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
						.notesize = GS_SMALL;
		}
		else if (Curr_grpsyl_p->grpcont == GC_REST) {
			Curr_grpsyl_p->grpsize  = GS_SMALL;
		}
		else {
			yyerror("no note specified for '?'");
		}
	}

	|
	T_TILDE opt_dir
	{
		if (Curr_grpsyl_p->grpcont != GC_NOTES) {
			yyerror("can't tie a rest or space");
		}
		else if (Curr_grpsyl_p->nnotes > 0) {
			struct NOTE *n_p;
			n_p = &(Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]);
			n_p->tie = YES;
			n_p->tiestyle = $1;
			n_p->tiedir = $2;
		}
		else {
			yyerror("no note specified for '~'");
		}
	}

	|
	T_EQUAL loc_variable
	{
		if (Curr_grpsyl_p->nnotes > 0) {
			switch (Curr_grpsyl_p->notelist
                                        [Curr_grpsyl_p->nnotes - 1].letter) {
			case PP_REST:
			case PP_SPACE:
			case PP_RPT:
			case PP_NO_PITCH:  /* this one not really possible */
				addsym($2, Curr_grpsyl_p->c, CT_GRPSYL);
				break;
			default:
				addsym($2, Curr_grpsyl_p->notelist
					[Curr_grpsyl_p->nnotes - 1].c, CT_NOTE);
				break;
			}
			var_valid();
		}
		else if (Curr_grpsyl_p->grpcont == GC_REST) {
			/* This should really never be hit anymore since
			 * chord-at-a-time code was added, but shouldn't
			 * hurt to leave it, just in case. */
			addsym($2, Curr_grpsyl_p->c, CT_GRPSYL);
			var_valid();
		}
		else {
			l_yyerror(Curr_filename, yylineno,
				"no note specified for location tag '%s'", $2);
		}
	}

	|
	T_HAT pitch opt_octave
	{
		/* this is for bend on a non-tablature staff */
		if (Doing_tab_staff == YES) {
			yyerror("^ bend not allowed on tablature staff; use quoted bend string");
		}
		else {
			add_slurto(Curr_grpsyl_p, $2, $3,
					Curr_grpsyl_p->nnotes - 1, L_NORMAL);
			if (Curr_grpsyl_p->nnotes > 0) {
				Curr_grpsyl_p->notelist
						[Curr_grpsyl_p->nnotes - 1]
						.is_bend = YES;
			}
		}
	}

	|
	T_HAT T_SLASH
	{
		/* this is for a small bend (1/4 step) on a non-tab staff */
		if (Doing_tab_staff == YES) {
			yyerror("^/ not allowed on tablature staff; use quoted bend string");
		}
		else if (Curr_grpsyl_p != 0 && Curr_grpsyl_p->nnotes > 0) {
			Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
					.smallbend = YES;
		}
	}

	|
	T_HS string
	{
		if (Curr_grpsyl_p->grpcont == GC_NOTES &&
					Curr_grpsyl_p->nnotes > 0) {
			if ((Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
					.headshape = get_shape_num($2 + 2))
					== HS_UNKNOWN) {
				l_yyerror(Curr_filename, yylineno,
					"'%s' is not a valid head shape name",
					ascii_str($2, YES, NO, TM_NONE));
			}
		}
		else {
			yyerror("no note specified for headshape");
		}
		FREE($2);
	}

	|
	slur_item
	;

note_opts:

	|
	note_opts extra_item
	;

notelist:
	some_notes opt_other_staff
	{
		$$ = $1 + $2;	/* total number of notes */
	};

opt_other_staff:
	/* empty */
	{
		$$ = 0;
	}

	|
	css_with some_notes T_PLACE
	{
		switch ($3) {
		case PL_ABOVE:
			Curr_grpsyl_p->stemto = CS_ABOVE;
			break;
		case PL_BELOW:
			Curr_grpsyl_p->stemto = CS_BELOW;
			break;
		default:
			yyerror("cross staff stem must be with 'above' or 'below'");
			Curr_grpsyl_p->stemto = CS_SAME;
		}
		if ($2 == 0) {
			/* Maybe this wouldn't really hurt to allow,
			 * but it's rather silly--why would user bother to go
			 * to the trouble of saying there are cross staff stem
			 * notes, but then not list any?
			 */
			yyerror("cross-staff stem note list is empty");
		}
		$$ = $2;
	};

css_with:	T_WITH
	{
		CSSused = YES;
		Curr_grpsyl_p->stemto_idx = Curr_grpsyl_p->nnotes;
	};

some_notes:
	{
		/* No notes.  If this is for the list of "normal" staff notes,
		 * and there are no "other" staff notes (for cross-staff stems),
		 * this means use the same notes as the last group. */
		$$ = 0;
	}

	|
	some_notes notedata note_opts
	{
		/* return number of notes in notelist */
		$$ = $1 + 1;
	};

notedata:	notedata1

	|
	T_LET_R
	{
		add_note(Curr_grpsyl_p, (int) PP_REST, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
	}

	|
	T_LET_S
	{
		/* temporarily stash uncompressibility (NO in this case)
		 * in the octave field */
		add_note(Curr_grpsyl_p, (int) PP_SPACE, 0, NO,
							0, NO, (char *) 0);
		if (Curr_grpsyl_p->is_meas == YES && User_meas_time == YES) {
			l_warning(Curr_filename, yylineno,
				"specifying time value on measure space is pointless; ignoring");
		}
	}

	|
	T_LET_U T_LET_S
	{
		/* temporarily stash uncompressibility (YES in this case)
		 * in the octave field */
		add_note(Curr_grpsyl_p, (int) PP_SPACE, 0, YES,
							0, NO, (char *) 0);
		if (Curr_grpsyl_p->is_meas == YES && User_meas_time == YES) {
			l_warning(Curr_filename, yylineno,
				"specifying time value on measure uncollapseable space is pointless; ignoring");
		}
	}

	|
	T_RPT
	{
		if (Curr_grpsyl_p->is_meas == YES) {
			if (User_meas_time == YES) {
				l_warning(Curr_filename, yylineno,
					"specifying time value on measure repeat is pointless; ignoring");
			}
		}
		else {
			yyerror("rpt can only be used with m");
		}
		add_note(Curr_grpsyl_p, (int) PP_RPT, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
	}

	|
	T_BARTYPE
	{
		if ($1 == SINGLEBAR) {
			/* lexer thinks this is 'bar'
			 * but is really 'b' 'a' 'r' */
			keyword_notes(yytext);
		}
		else if ($1 == ENDBAR) {
			/* lexer thinks this is 'endbar'
			 * but is really 'en' 'd' 'b' 'a' 'r' */
			add_note(Curr_grpsyl_p, (int) 'e', 'n',
				USE_DFLT_OCTAVE, 0, NO, (char *) 0);
			keyword_notes(yytext + 2);
		}
		else {
			yyerror("bar type not valid here");
		}
	}

	|
	T_GRACE
	{
		/* lexer thinks this is 'grace' but really 'g' 'r' 'a' 'c' 'e' */
		keyword_notes(yytext);
	}

	|
	T_ENDTYPE
	{
		if ($1 == ENDING_BARRED) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}

	|
	T_PARATYPE {
		if ($1 == J_RAGPARA) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}

	| 
	T_RANGELISTVAR
	{
		if ($1 == BRACELIST) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}

	|
	T_CLEF
	{
		if ($1 == BASS) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}

	|
	T_USEACCS
	{
		/* in chord-at-time mode, lex could mistake useaccs
		 * for a 6-note chord: us e a c c s  */
		keyword_notes(yytext);
	}

	|
	T_LPAREN notedata1 T_RPAREN
	{
		Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1].note_has_paren = YES;
	};

notedata1:	pitch opt_bend
	{
		if (Doing_tab_staff == YES) {
			add_note(Curr_grpsyl_p, $1, 0, NOFRET, 0, NO, $2);
		}
		else {
			add_note(Curr_grpsyl_p, $1, 0, USE_DFLT_OCTAVE, 0, NO, $2);
		}
	}

	|
	pitch T_LPAREN acc_symbol T_RPAREN opt_octave
	{
		if (Doing_tab_staff == YES) {
			l_warning(Curr_filename, yylineno,
				"parentheses around accidentals are extraneous on tab staffs; ignoring");
		}
		add_note(Curr_grpsyl_p, $1, $3, $5, 0, YES, (char *) 0);
	}

	|
	pitch T_LPAREN octave T_RPAREN opt_bend
	{
		if (Doing_tab_staff == NO) {
			l_warning(Curr_filename, yylineno,
				"parentheses around octave are extraneous; ignoring");
		}
		add_note(Curr_grpsyl_p, $1, 0, $3, 0, YES, $5);
	}

	|
	pitch T_LPAREN pitch
	{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, $1, 0,
			(Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE),
			0, NO, (char *) 0);
	}

	|
	pitch acc_symbol opt_ticks opt_octave opt_bend
	{
		add_note(Curr_grpsyl_p, $1, $2, $4, $3, NO, $5);
	}

	|
	pitch acc_symbol opt_ticks T_LPAREN octave T_RPAREN opt_bend
	{
		if (Doing_tab_staff == NO) {
			l_warning(Curr_filename, yylineno,
				"extraneous parentheses around octave; ignoring");
		}
		add_note(Curr_grpsyl_p, $1, $2, $5, $3, YES, $7);
	}

	|
	pitch acc_symbol opt_ticks T_LPAREN pitch
	{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, $1, $2,
			(Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE),
			$3, NO, (char *) 0);
	}

	|
	pitch T_TICKS opt_octave opt_bend
	{
		add_note(Curr_grpsyl_p, $1, 0, $3, $2, NO, $4);
	}

	|
	pitch T_TICKS T_LPAREN octave T_RPAREN opt_bend
	{
		add_note(Curr_grpsyl_p, $1, 0, $4, $2, YES, $6);
	}

	|
	pitch octave opt_accidental opt_bend
	{
		if (Doing_tab_staff == YES && $3 != '\0') {
			yyerror("accidental must be before fret number");
		}
		add_note(Curr_grpsyl_p, $1, $3, $2, 0, NO, $4);
	}

	|
	pitch octave T_LPAREN acc_symbol T_RPAREN
	{
		add_note(Curr_grpsyl_p, $1, $4, $2, 0, YES, (char *) 0);
	}

	|
	pitch octave T_LPAREN pitch
	{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, $1, 0, $2, 0, NO, (char *) 0);
	};

opt_bend:
	{
		$$ = (char *) 0;
	}

	|
	string
	{
		if (Doing_tab_staff == NO) {
			/* try to give helpful error message */
			if (get_shape_num($1 + 2) != HS_UNKNOWN) {
				yyerror("missing 'hs' before headshape string");
			}
			else if (strcmp($1 + 2, "full") == 0 || isdigit($1[2])) {
				yyerror("bend string not allowed on non-tablature staff; use ^");
			}
			else {
				yyerror("unexpected string");
			}
			$$ = (char *) 0;
		}
	};

pitch:	T_PITCH
	{
		/* return 'a' to 'g' value */
		$$ = (int) *yytext;
	};

opt_accidental:
	{
		/* no accidental */
		$$ = 0;
	}

	|
	acc_symbol
	{
	};

opt_octave:
	{
		/* no octave or fret designation */
		$$ = (Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE);
	}

	|
	octave
	{
	};

acc_symbol:
	T_SHARP
	{
		$$ = '#';
	}

	|
	T_AMPERSAND
	{
		$$ = '&';
	}

	|
	T_LET_N
	{
		/* natural */
		$$ = 'n';
	}

	|
	T_LET_X
	{
		/* double sharp */
		$$ = 'x';
	}

	|
	T_DBLFLAT
	{
		$$ = 'B';
	};

octave:
	num
	{
		$$ = $1;
	}

	|
	minuslist
	{
		/* we can't really fill in the actual octave yet, because
		 * it may be different on different staffs or voices, so
		 * we store the relative octave and fill in actual value
		 * later.
		 */
		/* this will be a negative number */
		$$ = $1;
	}

	|
	pluslist
	{
		/* we can't really fill in the actual octave yet, because
		 * it may be different on different staffs or voices, so
		 * we store the relative octave and fill in actual value
		 * later.
		 */
		$$ = USE_DFLT_OCTAVE + $1;
	};

minuslist:	T_DASH
	{
		$$ = -1;
	}

	|
	minuslist T_DASH
	{
		/* count up the number of minus signs */
		$$ = $1 - 1;
	};

pluslist:	T_PLUS
	{
		$$ = 1;
	}

	|
	pluslist T_PLUS
	{
		/* count up the number of plus signs */
		$$ = $1 + 1;
	};

slur_item:
	start_slur slurlist T_R_ANGLE opt_dir
	{
		set_slurdir(Curr_grpsyl_p, Curr_grpsyl_p->nnotes - 1, $4);
	}
	;

start_slur:
	T_L_ANGLE
	{
		Linetype = $1;
		begin_slur(Curr_grpsyl_p, Curr_grpsyl_p->nnotes - 1);
	};

slurlist:
	{
		/* empty list. Only allowed if only one note in following
		 * group. However, we don't know that yet, so marked pitch
		 * as 'U' for unknown */
		add_slurto(Curr_grpsyl_p, 'U', USE_DFLT_OCTAVE,
				Curr_grpsyl_p->nnotes - 1, Linetype);
	}

	|
	T_NOWHERE_SLIDE
	{
		add_slurto(Curr_grpsyl_p, 0, $1, Curr_grpsyl_p->nnotes - 1,
						Linetype);
	}

	|
	fullslurlist
	;

fullslurlist:
	slurnote

	|
	fullslurlist slurnote
	;

slurnote:
	pitch opt_octave
	{
		if (Doing_tab_staff == YES) {
			yyerror("cannot specify string inside <> on tab staffs");
		}
		else {
			add_slurto(Curr_grpsyl_p, $1, $2,
					Curr_grpsyl_p->nnotes - 1, Linetype);
		}
	};

grp_locsave:
	T_EQUAL loc_variable 
	{
		/* save address associated with entire group */
		if ( $2 != (char *) 0) {
			addsym($2, Curr_grpsyl_p->c, CT_GRPSYL);
			var_valid();
		}
	};

loc_variable:	T_LETTER 
	{
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_PITCH
	{
		/* a-g are usually pitches, but in this context, they
		 * are 1-character variable names. */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_M
	{
		/* usually m means measure, but here it is the variable m */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_R
	{
		/* usually rest, here variable r */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_S
	{
		/* usually space, here variable s */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_U
	{
		/* usually uncompressible, here variable u */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_N
	{
		/* usually natural, here variable n */
		$$ = strcpy(Stringbuff, yytext);
	}

	|
	T_LET_X
	{
		/* usually double sharp, here variable x */
		$$ = strcpy(Stringbuff, yytext);
	}

	| T_LVAR
	{
		/* longer variable -- one with more than one character name */
		$$ = strcpy(Stringbuff, yytext);
	};

ntuplet:	start_tuplet groupinfo rbrace opt_side num printtup opt_time T_SEMICOLON
	{
		end_tuplet($5, $7, $6, $4);
		Getting_tup_dur = NO;
	};

printtup:
	{
		/* nothing -- use default of when to print tuplet number/bracket */
		$$ = PT_DEFAULT;
	}

	|
	T_LET_N
	{
		/* don't print tuplet or bracket */
		$$ = PT_NEITHER;
	}

	|
	T_NUM
	{
		/* print number only */
		$$ = PT_NUMBER;
	}

	|
	T_LETTER
	{
		if (*yytext == 'y') {
			$$ = PT_BOTH;
		}
		else {
			yyerror("tuplet number/bracket qualifier must be y or n or num");
		}
	};


start_tuplet:
	T_LBRACE
	{
		begin_tuplet();
	};

rbrace:		T_RBRACE
	{
		Getting_tup_dur = YES;
	};

opt_side:
	{
		$$ = PL_UNKNOWN;
	}

	|
	T_PLACE
	{
		if (yylval.intval == PL_BETWEEN) {
			yyerror("between not allowed for tuplet side");
			$$ = PL_UNKNOWN;
		}
	};

opt_time:
	{
		/* optional time value is missing */
		$$.n = 0;
		$$.d = 1;
	}

	|
	T_COMMA time_item
	{
		$$ = $2;
	};

lyricinfo:	T_LYRICS lyric_id T_COLON lyricdata
	{
	};

lyric_id:	lyr_opt_place staff_list more_lyric_ids
	{
		Last_grpsyl_p = Curr_grpsyl_p = (struct GRPSYL *) 0;
	};

lyr_opt_place:	opt_place
	{
		/* If user does "all" without specifying a place,
		 * all() will call dflt_place() which was really intended
		 * to only handle STUFF. So we have to tell it we have lyrics,
		 * not a true STUFF. */
		set_stuff_type(-1);
	};

opt_place:
	{
		/* empty, use default */
		begin_range(PL_UNKNOWN);
		$$ = PL_UNKNOWN;
	}

	|
	T_PLACE
	{
		begin_range($1);
	};

staff_list:	staff_item

	|
	T_ALL
	{
		/* means goes above or below all the staffs, regardless if
		 * some happen to be invisible at the moment, so find top
		 * or bottom visible staff as appropriate,and save that
		 * away as the staff range */
		all();
	}

	|
	staff_list T_COMMA staff_item
	;

staff_item: num staff_range
	{
		/* if staff_range == 0, then only one staff specified, so
		 * use first staff number for both beginning and end of range */
		save_staff_range($1, ($2 == 0 ? $1 : $2) );
	};

staff_range:
	{
		/* empty */
		chk_range_type(NO);
		$$ = 0;
	}

	|
	T_DASH num
	{
		chk_range_type(NO);
		$$ = $2;
		if ($2 == 0) {
			yyerror("staff of 0 is illegal");
		}
	}

	| 
	T_AMPERSAND num
	{
		chk_range_type(YES);
		$$ = $2;
	};

more_lyric_ids:
	/* empty */
	
	|
	more_lyric_ids T_SEMICOLON T_PLACE { Place = $3; } staff_list
	;

verse_spec:
	verselist
	{
	}

	|
	T_PITCH
	{
		if (*yytext != 'c') {
			yyerror("verse must be 'c' or number(s)");
		}
		else {
			save_vno_range(0, 0);
		}
	};

verselist:
	verse_item
	{
	}

	|
	verselist T_COMMA verse_item
	{
	};

verse_item:
	num verse_range
	{
		/* if end of range is 0, not really a range, use first number
		 * for both beginning and end */
		lyr_verse($1, ($2 == 0 ? $1 : $2));
	};

verse_range:
	{
		/* empty */
		$$ = 0;
	}

	|
	T_DASH num
	{
		$$ = $2;
	};

alloc_lyr_grpsyl:
	{
		/* null token to allocate a GRPSYL for a lyric */
		Last_grpsyl_p = Curr_grpsyl_p;
		Curr_grpsyl_p = newGRPSYL(GS_SYLLABLE);
		if (Last_grpsyl_p == (struct GRPSYL *) 0) {
			Lyrics_p = Curr_grpsyl_p;
		}
	};

lyricdata:	lyr_tlist lyr_verse
	{
		/* If user didn't specify a place, fix that. */
		if (Place == PL_UNKNOWN) {
			Place = PL_BELOW;
		}
		/* copies of the lyrics info has been made for all staffs/verses
		 * at this point, so get rid of master copy */
		free_grpsyls(Lyrics_p);
	};

lyr_tlist:
	{
		/* empty -- need to derive times from music */
		Lyrics_p = derive_lyrtime();
	}

	|
	lyrics_timelist
	;

lyrics_timelist: lyr_time_item

	|
	lyrics_timelist lyr_time_item 
	;

lyr_time_item:	lyr_tm T_SEMICOLON
	;

lyr_tm:	alloc_lyr_grpsyl time_val opt_space
	{
		link_notegroup(Curr_grpsyl_p, Last_grpsyl_p);
		/* Save pointer to current GRPSYL in case the next
		 * group gets its time value based on this one,
		 * and there are additive times on this one. */
		Prev_grpsyl_p = Curr_grpsyl_p;
		/* If there are additive times, add those in.
		 * We don't need to make extra groups for lyrics, because we
		 * don't need to tie groups together--the lyrics time
		 * can hold any legal RATIONAL, even those that aren't
		 * specifiable with a single time value. */
		if (Extra_time_p != 0) {
			struct TIMELIST *timelist_p;
			for (timelist_p = Extra_time_p; timelist_p != 0;
						timelist_p = timelist_p->next) {
				Curr_grpsyl_p->fulltime = radd(Curr_grpsyl_p->fulltime, timelist_p->fulltime);
			}
		}
	}

	|
	lyrtuplet
	{
	};

lyrtuplet:	start_tuplet lyrics_timelist rbrace num opt_time
	{
		end_tuplet($4, $5, NO, PL_UNKNOWN);
		Getting_tup_dur = NO;
	};

opt_space:
	{
		/* empty, will fill in an actual syllable later */
		Curr_grpsyl_p->syl = (char *) 0;
	}

	|
	T_LET_S
	{
		/* space, not a lyric, so mark for later use */
		Curr_grpsyl_p->grpcont = GC_SPACE;
	};

lyr_verse:	verse_string

	|
	lyr_verse verse_string
	;

verse_string:
	string T_SEMICOLON
	{
		/* If no [verseno] is specified, use -1 as special flag
		 * to be resolved later to "one more than previous",
		 * or if there wasn't a previous, to verse 1.
		 */
		lyr_verse(-1, -1);
		proc_lyrics(Lyrics_p, $1);
	}

	|
	T_LBRACKET verse_spec T_RBRACKET string T_SEMICOLON
	{
		proc_lyrics(Lyrics_p, $4);
	};

os_directive:	barinfo bar_items
	{
		Currstruct_p = (struct MAINLL *) 0;
		/* If there are alternating time signature,
		 * add an implicit time signature SSV for the
		 * next time signature in the list.
		 */
		if (Alt_timesig_list != 0) {
			end_prev_context();
			Context = C_SCORE;
			Currstruct_p = newMAINLLstruct(S_SSV, -1);
			Currstruct_p->u.ssv_p->context = Context;
			if (Tsig_visibility == PTS_ALWAYS &&
					Next_alt_timesig != Alt_timesig_list) {
				/* If user wants alternating time signatures
				 * printed on every measure, if there is
				 * a multirest, we will print multiple
				 * time signatures there, and might have to
				 * wrap around to the beginning of the list.
				 * So make a copy of the entire list,
				 * starting from wherever we are now,
				 * wrapping around to the beginning,
				 * and ending just before where we are now.
				 * Most of the time,
				 * there probably won't be a multirest,
				 * so this will be a waste, but unfortunately,
				 * we don't know yet whether there will be
				 * one or not, so need to do the whole list
				 * just in case.
				 *
				 * Calculate length of the two pieces:
				 * from where we are to end, and from
				 * beginning to where we are.
				 */
				int remlength;
				int wraplength;
				remlength = strlen(Next_alt_timesig);
				wraplength = strlen(Alt_timesig_list)
						- remlength;
				/* need one more for terminator. */
				MALLOCA(char, Currstruct_p->u.ssv_p->timerep,
						remlength + wraplength + 1);

				/* copy remainder into beginning */
				strcpy(Currstruct_p->u.ssv_p->timerep,
						Next_alt_timesig);
				/* copy the wrap-around part of list,
				 * but move the TSR_ALTERNATING from the
				 * end of that part to between the two parts */
				Currstruct_p->u.ssv_p->timerep[remlength]
						= TSR_ALTERNATING;
				strncpy(Currstruct_p->u.ssv_p->timerep +
						remlength + 1, Alt_timesig_list,
						wraplength - 1);
				/* Add the terminator */
				Currstruct_p->u.ssv_p->timerep
						[remlength + wraplength]
						= TSR_END;

				assign_timesig(Currstruct_p, PTS_ALWAYS,
							&Next_alt_timesig);


				/* Make this new copy the new
				 * head of the list */
				Alt_timesig_list =
						Currstruct_p->u.ssv_p->timerep;
			}
			else {
				/* make a copy of the remaining alternating
				 * signatures and assign that */
				MALLOCA(char, Currstruct_p->u.ssv_p->timerep,
					strlen(Next_alt_timesig) + 1);
				strcpy(Currstruct_p->u.ssv_p->timerep,
							Next_alt_timesig);
				assign_timesig(Currstruct_p,
					(Tsig_visibility == PTS_ALWAYS ?
					PTS_ALWAYS : PTS_NEVER),
					&Next_alt_timesig);

				/* If we reached end of list, start over */
				if (Next_alt_timesig == 0) {
					Next_alt_timesig = Alt_timesig_list;
				}
			}

			asgnssv(Currstruct_p->u.ssv_p);
			end_prev_context();
			Context = C_MUSIC;
		}
	}

	|
	line

	|
	curve

	|
	T_SCOREFEED
	{
		struct MAINLL * mll_p;
		char *name;


		name = (yylval.intval == YES ? "newpage" : "newscore");

		(void) contextcheck(C_MUSIC | C_BLOCK, name);
		Currstruct_p = newMAINLLstruct(S_FEED, yylineno);
		Currstruct_p->u.feed_p->pagefeed = $1;

		/* make sure we're not in the middle of a measure. Go
		 * backwards in main list. If we hit STAFF before a BAR,
		 * then there is a problem.
		 */
		for (mll_p = Mainlltc_p; mll_p != (struct MAINLL *) 0;
					mll_p = mll_p->prev) {

			if (mll_p->str == S_BAR) {
				break;
			}
			else if (mll_p->str == S_STAFF) {
				l_yyerror(Curr_filename, yylineno,
					"%s not allowed in middle of measure",
					name);
				break;
			}
			else if (mll_p->str == S_FEED) {
				if (mll_p->prev != 0 &&
						mll_p->prev->str == S_BLOCKHEAD) {
					/* This is a feed following a block.
					 * If it was an implicit feed,
					 * we can get rid of it, because
					 * user now put an explicit one.
					 */
					if (mll_p->inputlineno == -1) {
						unlinkMAINLL(mll_p);
						FREE(mll_p);
					}
					/* We must be in block context,
					 * so no need to check farther back.
					 */
					break;
				}
				else {
					l_yyerror(Curr_filename, yylineno,
						"consecutive newscore/newpage not allowed");
				}
				break;
			}
		}
		insertMAINLL(Currstruct_p, Mainlltc_p);

		/* If this is inside a block, we need to create a new block
		 * and feed after it.
		 */
		if (Context == C_BLOCK) {
			CALLOC(BLOCKHEAD, Currblock_p, 1);
			set_win_coord(Currblock_p->c);
			Next_print_link_p_p = &(Currblock_p->printdata_p);
			insertMAINLL(newMAINLLstruct(S_BLOCKHEAD, yylineno),
						Mainlltc_p);
			Mainlltc_p->u.blockhead_p = Currblock_p;
			insertMAINLL(newMAINLLstruct(S_FEED, -1), Mainlltc_p);
		}

	}
	opt_temp_margin
	{
		Currstruct_p = (struct MAINLL *) 0;
	}

	|
	printcmd
	;

opt_temp_margin:
	/* empty */

	|
	opt_temp_margin temp_margin
	;

temp_margin:	T_FNUMVAR opt_equals floatnum
	{
		$3 = adjust2inches($3);	/* in case we are in centimeter mode */
		if ($1 == RIGHTMARGIN) {
			if (Currstruct_p->u.feed_p->rightmargin >= 0) {
				l_warning(Curr_filename, yylineno,
					"rightmargin specified more than once, using last instance");
				/* fall through to override the previous */
			}
			Currstruct_p->u.feed_p->rightmargin = $3;
		}
		else if ($1 == LEFTMARGIN) {
			if (Currstruct_p->u.feed_p->leftmargin >= 0) {
				l_warning(Curr_filename, yylineno,
					"leftmargin specified more than once, using last instance");
				/* fall through to override the previous */
			}
			Currstruct_p->u.feed_p->leftmargin = $3;
		}
		else {
			yyerror("unexpected parameter; only 'leftmargin' or 'rightmargin' allowed here");
		}
		chkmargin(Score.topmargin, Score.botmargin,
				Currstruct_p->u.feed_p->leftmargin,
				Currstruct_p->u.feed_p->rightmargin);
	};
	
opt_equals:	/* optional equal sign */
	/* empty */
	|
	T_EQUAL
	;

opt_semi:
	/* optional semi-colon */
	|
	T_SEMICOLON
	;
	
directive:
	stuff_info

	|
	roll_cmd
	;

barinfo:	linetype precbartype barline
	{

		/* build a BAR struct and add it to the main list */
		if (contextcheck(C_MUSIC, "bar") == YES) {

			Currstruct_p = add_bar($3, $1, Endingloc, Mainlltc_p);
			Got_ending = NO;
		}
		if ($3 == RESTART && Endingloc != NOITEM) {
			yyerror("restart cannot be used inside an ending");
		}

		/* User can specify bar type to use on preceeding staff when
		 * a repeatstart gets moved to the next scores's pseudo-bar */
		if ($2 != -1) {
			if ($3 != REPEATSTART) {
				l_warning(Curr_filename, yylineno,
					"bar type for preceeding score only allowed on repeatstart; ignoring");
			}
			else if (Currstruct_p != 0 && Currstruct_p->u.bar_p != 0) {
				Currstruct_p->u.bar_p->precbartype = $2;
				if ($2 != SINGLEBAR && $2 != DOUBLEBAR &&
						(Linetype == L_DASHED ||
						Linetype == L_DOTTED)) {
					l_warning(Curr_filename, yylineno,
						"dashed/dotted only allowed on bar or dblbar; ignoring");
					Linetype = L_NORMAL;
				}
				Currstruct_p->u.bar_p->linetype = Linetype;
			}
		}
		else {
			/* Default is single bar. It could be argued that in
			 * the case of a key change on this bar, it really
		 	 * should be a dblbar, but user can force that
			 * if they want it. */
			if (Currstruct_p != 0 && Currstruct_p->u.bar_p != 0) {
				Currstruct_p->u.bar_p->precbartype = SINGLEBAR;
			}
		}
		/* Placement phase indirectly uses Doing_tab_staff when it
		 * calls notecomp, so clear it in case we don't get a
		 * non-tab staff input before the end of file.
		 * We could probably get by just doing after parsing is done,
		 * but this seems cleaner, and cheap enough.
		 */
		Doing_tab_staff = NO;
	};

precbartype:
	/* empty */
	{
		$$ = -1;	/* special value to mean "none specified" */
		Linetype = L_NORMAL;
	}

	|
	T_LPAREN linetype T_BARTYPE T_RPAREN
	{
		/* Parens wouldn't be strictly necessary to be able to parse,
		 * but seem useful semantically to make it clear this is
		 * optional, only applying when repeatstart is moved to
		 * next score's pseudo bar. */
		$$ = $3;
		if ($2 == L_DOTTED || $2 == L_DASHED) {
			if ($3 != SINGLEBAR && $3 != DOUBLEBAR) {
				l_warning(Curr_filename, yylineno,
					"Only bar or dblbar can be dashed or dotted; ignoring");
			}
			else {
				Linetype = $2;
			}
		}
		else if ($2 != L_NORMAL) {
			l_warning(Curr_filename, yylineno,
					"Bar line type modifier can only can be dashed or dotted; ignoring");
		}
	};

barline:	bartype emptymeas_input
	{
		/* Call proc_emptymeas() to free up the last buffer, if any,
		 * and return back to normal input stream. */
		proc_emptymeas(MAXSTAFFS+1);
		$$ = $1;
	};

bartype:	T_BARTYPE
	{
		/* If there was any empty voices in the input for this
		 * measure which have emptymeas parameter set,
		 * proc_emptymeas() will arrange to create just-in-time
		 * virtual input for it that will get parsed via
		 * the emptymeas_input rule.
		 * We pass arg of 1 to tell it to start looking from
		 * staff 1. Subsequent calls to proc_emptymeas(), if any,
		 * will generate input for additional voices, until
		 * no more are needed for this measure.
		 */
		if ($1 != RESTART) {
			EM_staff = proc_emptymeas(1);
		}
		else {
			/* don't create any music before a restart */
			EM_staff = MAXSTAFFS + 1;
		}
		$$ = $1;
	};

emptymeas_input:	/* empty */
	{
	}

	|
	emptymeas_input T_EM_BEGIN music_input T_EM_END
	{
		/* See if there is any more virtual input needed */
		EM_staff = proc_emptymeas(EM_staff);
	};

bar_items:

	|
	bar_items bar_opt
	;

bar_opt:
	padspec
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->padding += $1 * STEPSIZE;
		}
	}

	|
	T_EQUAL loc_variable
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			/* fill in location info */
			if ($2 != (char *) 0) {
				addsym($2, Currstruct_p->u.bar_p->c, CT_BAR);
			}
		}
	}

	|
	ending_info
	{
		if (Got_ending == YES) {
			yyerror("Only one ending allowed per bar");
		}
		Got_ending = YES;

		if (Currstruct_p != (struct MAINLL *) 0) {
			if (Currstruct_p->u.bar_p->bartype == RESTART) {
				yyerror("ending not allowed on restart");
			}
			/* fill in ending label if any */
			Currstruct_p->u.bar_p->endinglabel = $1;
			Currstruct_p->u.bar_p->endingloc = Endingloc;
			ped_endings(Endingloc);

			/* for next time around, figure out what endingloc will
			 * be if user doesn't specify something different */
			switch (Endingloc) {
			case STARTITEM:
				Endingloc = INITEM;
				break;
			case ENDITEM:
				Endingloc = NOITEM;
				break;
			default:
				break;
			}
		}
	}

	|
	T_HIDECHANGES
	{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.bar_p->bartype == RESTART) {
				l_warning(Curr_filename, yylineno,
					"hidechanges not allowed on restart; ignoring");
			}
			else {
				Currstruct_p->u.bar_p->hidechanges = YES;
			}
		}
	}

	|
	T_MNUM T_EQUAL num {
		if (Currstruct_p != 0) {
			set_mnum(Currstruct_p->u.bar_p, $3);
		}
	}

	|
	T_LET T_EQUAL string {
		/* +2 to skip font/size */
		init_reh(-1, $3 + 2, Currstruct_p);
	}

	|
	T_NUM T_EQUAL num {
		init_reh($3, (char *)0, Currstruct_p);
	}

	|
	T_REHEARSAL chk_dup_reh font_family title_font opt_size rehearsal_mark opt_dist
	{
		if (Currstruct_p != 0 && Currstruct_p->str == S_BAR) {
			if (Currstruct_p->u.bar_p->reh_string != 0) {
				/* This indicates user specified multiple
				 * rehearsal strings. We would have already
				 * given a warning message. */
				FREE(Currstruct_p->u.bar_p->reh_string);
				Currstruct_p->u.bar_p->reh_string = 0;
			}
			set_reh_string(Currstruct_p->u.bar_p, $3, $4, $5, $6);
			Currstruct_p->u.bar_p->dist = Dist;
			Currstruct_p->u.bar_p->dist_usage = Dist_usage;
		}
	};

ending_info:
	T_ENDING string
	{
		$$ = $2;
		(void) fix_string($2, FONT_TR, DFLT_SIZE, Curr_filename, yylineno);
		Endingloc = STARTITEM;
	}

	|
	T_ENDENDING
	{
		if (Endingloc == NOITEM) {
			l_warning(Curr_filename, yylineno,
				"no ending in progress; extraneous 'endending' ignored");
		}
		else {
			Endingloc = ENDITEM;
		}
		$$ = (char *) 0;
	};

chk_dup_reh:
	{
		/* null token to check for more than one rehearsal mark
		 * on one BAR */
		if (Currstruct_p != (struct MAINLL *) 0 &&
				Currstruct_p->u.bar_p->reh_type != REH_NONE) {
			l_warning(Curr_filename, yylineno,
				"only one rehearsal mark allowed per bar; using last");
		}
	};

rehearsal_mark:	string
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_STRING;
		}
	}

	|
	T_MNUM
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_MNUM;
		}
		$$ = (char *) 0;
	}

	|
	T_NUM
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_NUM;
		}
		$$ = (char *) 0;
	}

	|
	T_LET
	{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_LET;
		}
		$$ = (char *) 0;
	};

location:	T_LPAREN alloc_loc xcoord T_COMMA ycoord T_RPAREN
	{
		$$ = $2;
	};

alloc_loc:
	{
		/* null token to allocate an INPCOORD */
		CALLOC(INPCOORD, Curr_loc_info_p, 1);
		$$ = Curr_loc_info_p;
		/* set to positive in case input starts with an absolute
		 * x coordinate */
		Plus_minus = 1;
	};

xcoord:	xterm xoffsetlist
	{
		/* reset sign to positive in case y coordinate is absolute */
		Plus_minus = 1;
	};

xterm:		xoffsetval

	|
	variable T_DOT east_or_west
	{
		/* only 1 of these allowed per location */
		if (Curr_loc_info_p->hor_p != (float *) 0) {
			pfatal("only one horizontal location variable allowed per coordinate");
		}
		/* fill in the coord address and type */
		Curr_loc_info_p->hor_p = symval($1, &(Curr_loc_info_p->hor_p));
		Curr_loc_info_p->htype = $3;
	}

	|
	xoffset
	;


east_or_west:	T_LETTER
	{
		/* only valid letters is w */
		if (*yytext != 'w') {
			yyerror("invalid direction: must be x, w, or e");
		}
		$$ = AW;
	}
	|
	T_LET_X
	{
		$$ = AX;
	}

	|
	T_PITCH
	{
		/* only valid value is e */
		if (*yytext != 'e') {
			yyerror("invalid direction: must be x, w, or e");
		}
		$$ = AE;
	};

xoffsetlist:

	| xoffsetlist xoffset
	;

xoffset:	plus_or_minus xoffsetval
	;

plus_or_minus:	T_PLUS
	{
		$$ = Plus_minus = 1;
	}

	|
	T_DASH
	{
		$$ = Plus_minus = -1;
	};

xoffsetval:	floatnum
	{
		Curr_loc_info_p->hsteps += $1 * (float) Plus_minus;
	}

	|
	T_TIME floatnum
	{
		if (Curr_loc_info_p->hor_p == (float *) 0) {
			yyerror("must specify location before giving time offset");
		}
		/* add or subtract time value from horizontal */
		if (Plus_minus == 1) {
			Curr_loc_info_p->counts += $2;
		}
		else {
			Curr_loc_info_p->counts -= $2;
		}
	};

ycoord:	yterm yoffsetlist
	;

yterm:		yoffsetval

	|
	variable T_DOT north_or_south
	{
		/* only one of these allowed per location */
		if (Curr_loc_info_p->vert_p != (float *) 0) {
			pfatal("only one vertical location variable allowed per coordinate");
		}
		/* fill in the coord address and type */
		Curr_loc_info_p->vert_p = symval($1, &(Curr_loc_info_p->vert_p));
		Curr_loc_info_p->vtype = $3;

	};

north_or_south: T_LET_N
	{
		$$ = AN;
	}

	|
	T_LET_S
	{
		$$ = AS;
	}

	|
	T_LETTER
	{
		/* only valid value here is y */
		if (*yytext != 'y') {
			yyerror("invalid direction: must be y, n, or s");
		}
		$$ = AY;
	};

yoffsetlist:

	|
	yoffsetlist yoffset
	;

yoffset:	plus_or_minus yoffsetval
	{
	};

yoffsetval:
	floatnum
	{
		/* this is number of steps up (or down if negative) */
		Curr_loc_info_p->vsteps += $1 * (float) Plus_minus;
	};

floatnum:	num
	{
		$$ = (float) $1;
	}

	| num T_DOT opt_decimal_part
	{
		sprintf(Tmpbuff,"%d.%s", $1, $3);
		$$ = (float) atof(Tmpbuff);
	}

	| T_DOT numstr
	{
		sprintf(Tmpbuff,"0.%s", $2);
		$$ = (float) atof(Tmpbuff);
	};

opt_decimal_part:
	{
		/* no decimal fraction part of float number */
		$$ = "";
	}

	|
	numstr
	;

numstr:	T_NUMBER
	{
		$$ = yytext;
	};

variable:	loc_variable
	;

line:		linetype T_LINE line_alloc location T_TO location opt_line_str
	{
		Currstruct_p->u.line_p->linetype = $1;
		Currstruct_p->u.line_p->start = *($4);
		Currstruct_p->u.line_p->end = *($6);
		rep_inpcoord($4, &(Currstruct_p->u.line_p->start));
		rep_inpcoord($6, &(Currstruct_p->u.line_p->end));
		Currstruct_p->u.line_p->string = $7;

		/* copies of the location info went into the LINE struct,
		 * so we can free the original copy of the information */
		if ($4) {
			FREE($4);
		}
		if ($6) {
			FREE($6);
		}
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p = (struct MAINLL *) 0;
	};

line_alloc:
	{
		/* null token to cause allocation of LINE struct */
		(void) contextcheck(C_MUSIC, "line");
		Currstruct_p = newMAINLLstruct(S_LINE, yylineno);
	};

linetype:
	{
		$$ = L_NORMAL;
	}

	|
	T_LINETYPE
	;

opt_line_str:
	/* empty */
	{
		$$ = 0;
	}

	|
	T_WITH tfont title_size string {
		$$ = fix_string($4, Titlefont, Titlesize, Curr_filename, yylineno);
	}
	;

curve:		linetype T_CURVE curve_alloc curve_loc T_TO curve_loc curve_tail
	{
		if ($1 == L_WAVY) {
			l_warning(Curr_filename, yylineno,
						"wavy curve not allowed; using normal curve");
			$1 = L_NORMAL;
		}
		Currstruct_p->u.curve_p->curvetype = $1;
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p = (struct MAINLL *) 0;
	};

curve_alloc:
	{
		(void) contextcheck(C_MUSIC, "curve");
		Currstruct_p = newMAINLLstruct(S_CURVE, yylineno);

		/* get space for a list of locations and mark it as empty */
		Item_count = 0;
		Currstruct_p->u.curve_p->ncoord = (short) Item_count;
		Currstruct_p->u.curve_p->nbulge = (short) Item_count;
		MALLOC(INPCOORD, Currstruct_p->u.curve_p->coordlist, ITEMS);
		Max_items = ITEMS;
	};

curve_tail:
	T_TO loc_list

	|
	T_BULGE bulgelist
	;

loc_list:	curve_loc

	|
	loc_list T_TO curve_loc
	;

bulgelist:
	bulge_value
	
	|
	bulgelist T_COMMA bulge_value
	;

bulge_value:
	opt_minus floatnum
	{
		/* get enough space to store some values if don't have enough */
		if (Currstruct_p->u.curve_p->nbulge == 0) {
			MALLOCA(float, Currstruct_p->u.curve_p->bulgelist, ITEMS);
		}
		else if ( (Currstruct_p->u.curve_p->nbulge % ITEMS) == 0) {
			REALLOCA(float, Currstruct_p->u.curve_p->bulgelist,
				Currstruct_p->u.curve_p->nbulge + ITEMS);
		}

		/* fill in the value and increment the count of how many */
		Currstruct_p->u.curve_p->bulgelist[Currstruct_p->u.curve_p->nbulge] = $1 * $2;
		(Currstruct_p->u.curve_p->nbulge)++;
	};

curve_loc:	location
	{
		struct INPCOORD *curve_inpcoord_p;

		/* If ran off end of list, make list bigger.
		 * We cannot use REALLOC here, since we need to update
		 * the tag reference information on each individual array
		 * element, so we alloc all new space, copy the existing data,
		 * and update the tag references, then free the old space,
		 * and redirect the pointer to the new space. */
		if (Currstruct_p->u.curve_p->ncoord >= Max_items) {
			struct INPCOORD *newlist_p;
			int n;
			Max_items += ITEMS;
			MALLOC(INPCOORD, newlist_p, Max_items);
			for (n = 0; n < Currstruct_p->u.curve_p->ncoord; n++) {
				newlist_p[n] = Currstruct_p->u.curve_p->coordlist[n];
				rep_inpcoord(
				    &(Currstruct_p->u.curve_p->coordlist[n]),
				    &(newlist_p[n]));
			}
			FREE(Currstruct_p->u.curve_p->coordlist);
			Currstruct_p->u.curve_p->coordlist = newlist_p;
		}

		/* Add this entry to the end of the list, update the
		 * tag reference to point to this permanent copy rather
		 * the temporary one we will be deleting, and update the
		 * count of how many elements in the list. */
		curve_inpcoord_p = &(Currstruct_p->u.curve_p->coordlist
					[Currstruct_p->u.curve_p->ncoord]);
		*curve_inpcoord_p = *($1);
		rep_inpcoord($1, curve_inpcoord_p);
		(Currstruct_p->u.curve_p->ncoord)++;
		FREE($1);
	};

stuff_info: stuff_header T_COLON stufflist
	{
		attach_stuff();
		/* so reset flag */
		Defining_multiple = NO;
	};

stuff_header:
	stuff_id opt_size opt_modifier opt_place staff_list opt_dist opt_align
	{
		chk_stuff_header($2, $3, $4, Dist_usage);
		add_to_sv_list();
	}

	|
	linetype T_PHRASE opt_size opt_place staff_list opt_dist
	{
		/* Separate rule for phrase because it has the linetype
		 * modifier in front, unlike any other STUFF. The size,
		 * and dist are not really allowed, but we match them
		 * if they are there in order to give more clear error
		 * message than just "syntax error," since there was
		 * already code elsewhere to check for these errors.
		 */
		set_stuff_type(ST_PHRASE);
		chk_stuff_header($3, $1, $4, Dist_usage);
		add_to_sv_list();
		Aligntag = NOALIGNTAG;
	}
	|
	midi midi_staff
	{
		/* need a separate rule for midi, because it can include
		 * a voice as well as staff. It also cannot have the chord,
		 * size or place, although those are checked elsewhere anyway */
		chk_stuff_header(-1, NO, PL_UNKNOWN, SD_NONE);
		Aligntag = NOALIGNTAG;
	};

midi:
	T_MIDI
	{
		set_stuff_type(ST_MIDI);
		begin_range(PL_UNKNOWN);
	};

midi_staff:
	voice_id

	|
	T_ALL
	{
		all();
		add_to_sv_list();
	};

stuff_id:
	stuff_type
	{
		/* dummy token for sole purpose of saving away current
		 * stuff type. We have to do this in order to have it in
		 * time to figure out the default place for "all" */
		set_stuff_type($1);
	};

stuff_type:
	font_family T_FONT
	{
		Curr_family = $1;

		switch ($2) {
		case FONT_TR:
			$$ = ST_ROM;
			break;
		case FONT_TI:
			$$ = ST_ITAL;
			break;
		case FONT_TB:
			$$ = ST_BOLD;
			break;
		case FONT_TX:
			$$ = ST_BOLDITAL;
			break;
		default:
			pfatal("unknown font");
			break;
		}
	}

	|
	T_L_ANGLE
	{
		if ($1 != L_NORMAL) {
			/* user used 'dotted' or 'dashed' which applies only
			 * to use of T_L_ANGLE for slurs. Use generic error
			 * message so looks like any other similar error. */
			yyerror("syntax error");
		}
		$$ = ST_CRESC;
	}

	|
	T_R_ANGLE
	{
		$$ = ST_DECRESC;
	}

	|
	T_PEDAL
	{
		$$ = ST_PEDAL;
	}

	|
	T_MUSSYM
	{
		$$ = ST_MUSSYM;
	}

	|
	T_OCTAVE
	{
		$$ = ST_OCTAVE;
	}
	;

opt_size:
	{
		/* use -1 to indicate user didn't specify so must use default */
		$$ = -1;
	}

	|
	T_LPAREN num T_RPAREN
	{
		$$ = $2;
	};

opt_modifier:
	{
		$$ = TM_NONE;
	}

	|
	T_MODIFIER
	{
		$$ = $1;
	};

opt_dist:
	{
		Dist = 0.0;
		Dist_usage = SD_NONE;
	}

	|
	T_FNUMVAR stuff_dist
	{
		if ($1 != DIST) {
			yyerror("unexpected parameter name instead of 'dist'");
		}
		else {
			if (frangecheck($2, -MAXDIST, MAXDIST, "dist") == YES) {
				Dist = $2;
			}
		}
	};

stuff_dist:
	floatnum opt_exclam
	{
		$$ = $1;
	}

	|
	T_DASH floatnum opt_exclam
	{
		if (Dist_usage != SD_FORCE) {
			l_warning(Curr_filename, yylineno,
				"negative dist requires !, setting to 0");
			$$ = 0.0;
		}
		else {
			$$ = - $2;
		}
	};

opt_exclam:
	{
		Dist_usage = SD_MIN;
	}

	|
	T_EXCLAM
	{
		Dist_usage = SD_FORCE;
	};

opt_align:
	{
		Aligntag = NOALIGNTAG;
	}

	|
	T_ALIGN num
	{
		if (rangecheck($2, MINALIGNTAG, MAXALIGNTAG, "align tag number") == YES) {
			Aligntag = $2;
		}
		else {
			Aligntag = NOALIGNTAG;
		}
	};
	
stufflist:
	stuff_item

	|
	stufflist stuff_item
	;

stuff_item:
	beat_offset steps_offset opt_string til_clause T_SEMICOLON
	{
		add_stuff_item($1, $2, Gracebackup, $3, Til_bars, Til_offset,
						Dist, Dist_usage, Aligntag);
	};

beat_offset:
	floatnum gracebackup
	{
		Til_bars = 0;
		Til_offset = 0.0;
		$$ = $1;
	};

gracebackup:
	{
		Gracebackup = 0;
	}

	|
	T_LPAREN T_DASH num T_RPAREN
	{
		Gracebackup = $3;
	};

steps_offset:
	{
		$$ = 0.0;
	}

	|
	T_LBRACKET plus_or_minus floatnum T_RBRACKET
	{
		$$ = $3 * (float) $2;
	};

opt_string:
	{
		$$ = (char *) 0;
	}

	|
	T_STAR
	{
		if (get_stuff_type() != ST_PEDAL) {
			yyerror("'*' only valid with pedal");
		}
		MALLOCA(char, $$, 12);
		$$[0] = FONT_MUSIC;
		$$[1] = DFLT_SIZE;
		sprintf($$ + 2, "\\(endped)");
	}

	|
	string
	{
		$$ = $1;
	};

til_clause:
	{
	}

	|
	T_TIL til_val
	{
	};

til_val:
	floatnum til_suffix
	{
		/** if there is a non-empty til_suffix, the floatnum better
		 ** have been an integer, and is really the number
		 * of bars. **/
		if ($2 == YES) {
			Til_bars = (int) $1;
			if ($1 - Til_bars != 0.0) {
				yyerror("number of measures in til clause must be a whole number");
			}
		}
		else {
			/* only a number, that means it was really a beat offset */
			Til_offset = $1;
		}
	};

til_suffix:
	{
		$$ = NO;
	}

	|
	T_LET_M til_offset
	{
		$$ = YES;
	};

til_offset:
	{
		Til_offset = 0.0;
	}

	|
	T_PLUS floatnum
	{
		Til_offset = $2;
	};

roll_cmd:	T_ROLL opt_roll_dir roll_extent T_COLON roll_offsetlist
	;

opt_roll_dir:
	{
		newROLLINFO();
	}
	
	|
	T_UP
	{
		newROLLINFO();
		setrolldir(UP);
	}

	|
	T_DOWN
	{
		newROLLINFO();
		setrolldir(DOWN);
	};

roll_extent:
	roll_start roll_end
	;

roll_start:
	staffnum
	{
		/* save staffnum and voice 1 for start of roll */
		rollparam($1, 1, -1, -1);
	}

	|
	staffnum voicenum
	{
		/* save staff and voice for start of roll */
		rollparam($1, $2, -1, -1);
	} ;

roll_end:

	|
	T_TO staffnum
	{
		/* save staffnum, voice 1 for end of roll */
		rollparam(-1, -1, $2, 1);
	}

	|
	T_TO staffnum voicenum
	{
		/* save staffnum and voice for end of roll */
		rollparam(-1, -1, $2, $3);
	};

roll_offsetlist:
	floatnum T_SEMICOLON
	{
		rolloffset( (double) $1);
	}

	|
	roll_offsetlist floatnum T_SEMICOLON
	{	
		rolloffset( (double) $2);
	};

printcmd:	printtype opt_loc string
	{
		if (Curr_family == FAMILY_DFLT) {
			Curr_family = Score.fontfamily;
		}
		Extra = extra_needed(Curr_family + Curr_font, Curr_size, $3);
		proc_printcmd($1, Curr_loc_info_p, $3, Curr_family + Curr_font,
			Curr_size, $2, NO, ($2 == YES ? Extra : (double) 0.0));
	}

	|
	postscript opt_loc string
	{
		if ($2 == YES) {
			warning("nl ignored for postscript");
		}
		proc_printcmd(J_NONE, Curr_loc_info_p, $3, FONT_UNKNOWN,
			DFLT_SIZE, NO, YES, (double) 0.0);
		end_raw();
	};

postscript:
	T_POSTSCRIPT
	{
		begin_raw();
	};

printtype:	T_PRINTTYPE
	;

opt_loc:
	{
		/* this means print at default location */
		Curr_loc_info_p = (struct INPCOORD *) 0;
		$$ = NO;
	}

	|
	T_NL
	{
		Curr_loc_info_p = (struct INPCOORD *) 0;
		$$ = YES;
	}

	|
	location
	{
		$$ = NO;
	};


%%

#include <stdio.h>


/* print error message along with current filename and line number */

int
yyerror(msg)

char *msg;

{
	if (strncmp(msg, "syntax error", 12) == 0 ||
			strncmp(msg, "parse error", 11) == 0) {
		/* beef up yacc's default syntax error message */
		if (*yytext == '\n') {
			l_yyerror(Curr_filename, yylineno,
				"error detected at end of line (maybe missing semicolon or other required element?)");
		}
		else if (is_internal_token(yytext) == YES) { 
			/* We add some funny tokens for our internal use
			 * when doing emptymeas parameter expansion.
			 * It is conceivable that there could be
			 * some (rare) scenarios where a user error
			 * in their value for that parameter would cause the
			 * parser to detect the error when it got to our
			 * funny token, and it would be very confusing to the
			 * user if we printed that out as the offending token,
			 * since they didn't input it, and it is chosen
			 * specifically to be strange looking.
			 */
			l_yyerror(Curr_filename, yylineno,
				"error detected during internal emptymeas parameter processing");
		}
		else {
			l_yyerror(Curr_filename, yylineno,
				"error detected at '%s'", yytext);
		}
	}
	else {
		l_yyerror(Curr_filename, yylineno, "%s", msg);
	}

	/* return something to keep lint happy */
	return(0);
}


/* if argument is surrounded by double quotes, remove them and return the
 * unquoted string. Actually just NULL out the end quote and return a
 * pointer to beyond the leading quote */

static char *
stripquotes(string)

char *string;

{
	char *p;

	p = string + strlen(string) - 1;
	if (*p == '"') {
		*p = '\0';
	}
	return( *string == '"' ? string + 1 : string);
}


/* make sure each header/footer type specified only once */

static void
chkdup_headfoot(flag, which)

int flag;	/* if non-zero, had gotten this head/foot before */
char *which;	/* header, footer, header2 or footer2 */

{
	if (flag != 0) {
		l_yyerror(Curr_filename, yylineno,
				"%s context specified more than once", which);
	}
}


/* when entering new context, need to save away everything about current one */

static void
end_prev_context()

{
	struct MAINLL *place_p;		/* where to insert SSV in main list */


	/* assume we will place at the end of main list. Later we'll set
	 * the right value if that assumption is wrong */
	place_p = Mainlltc_p;

	/* If current main list item hasn't been added to list yet,
	 * do that now. */
	if (Currstruct_p != (struct MAINLL *) 0 && Currstruct_p != place_p) {

		if (Currstruct_p->str == S_SSV) {

			if (List_of_staffs_p != (struct MAINLL *) 0) {
				yyerror("can't change context inside a measure");
				Currstruct_p = (struct MAINLL *) 0;
				return;
			}

			/* if is an SSV struct, we also need to set the
			 * correct values at this point, because the
			 * parser needs to know current values of some
			 * things (like font and point size) */
			asgnssv(Currstruct_p->u.ssv_p);

			/* if this SSV is after a FEED, we have to move it
			 * before the FEED in order to follow the rules of
			 * what comes when on the main list */
			for (   ; place_p != (struct MAINLL *) 0 &&
					place_p->str == S_FEED;
					place_p = place_p->prev) {
			}
		}

		insertMAINLL(Currstruct_p, place_p);
	}

	if (Currstruct_p != 0 && Currstruct_p->str == S_SSV) {
		/* memorize or recall beamstyle and timeunit if appropriate */
		remember_tsig_params(Currstruct_p);

		/* If user specified multiple staffs/voices, clone the
		 * necessary SSV nows */
		if (Currstruct_p->u.ssv_p->context == C_STAFF) {
			clone_staff_ssvs(place_p);
			free_rlists();
		}
		else if (Currstruct_p->u.ssv_p->context == C_VOICE) {
			clone_voice_ssvs(place_p);
			free_rlists();
		}
		Defining_multiple = NO;
	}

	if (Curr_usym_p != 0) {
		define_usym();
		Curr_usym_p = 0;
	}

	Curr_grpsyl_p = (struct GRPSYL *) 0;
	Currblock_p = (struct BLOCKHEAD *) 0;
	set_win_coord(0);
}


/* If user gave a list of staffs for a "staff" context, clone copies of
 * the SSV that we made for the first on the list for the rest of the list. */

static void
clone_staff_ssvs(insert_p)

struct MAINLL *insert_p;
{
	struct SVRANGELIST *svr_p;
	struct RANGELIST *sr_p;		/* list of staff ranges */
	struct SSV *clone_src_p;	/* the SSV to clone */
	int s;				/* staff number */

	clone_src_p = Currstruct_p->u.ssv_p;
	for (svr_p = Svrangelist_p; svr_p != 0; svr_p = svr_p->next) {
		for (sr_p = svr_p->stafflist_p; sr_p != 0; sr_p = sr_p->next) { 
			for (s = sr_p->begin; s <= sr_p->end && s <= Score.staffs; s++) {
				if (clone_src_p->staffno == s) {
					/* skip past the clone source */
					continue;
				}
				clone1ssv(clone_src_p, insert_p, s, clone_src_p->voiceno);
			}
		}
	}
}


/* If user gave a list of staffs and/or voices  for a "voice" context,
 * clone copies of the SSV that we made for the first on the list
 * for the rest of the list. */

static void
clone_voice_ssvs(insert_p)

struct MAINLL *insert_p;
{
	struct SVRANGELIST *svr_p;
	struct RANGELIST *sr_p;		/* list of staff ranges */
	struct RANGELIST *vr_p;		/* list of voice ranges */
	struct SSV *clone_src_p;	/* the SSV to clone */
	int s;				/* staff number */
	int v;				/* voice number */
	int numvoices;			/* how many voices on current staff */

	clone_src_p = Currstruct_p->u.ssv_p;
	for (svr_p = Svrangelist_p; svr_p != 0; svr_p = svr_p->next) {
	    for (sr_p = svr_p->stafflist_p; sr_p != 0; sr_p = sr_p->next) { 
		for (s = sr_p->begin; s <= sr_p->end && s <= Score.staffs; s++) {
		    numvoices = vscheme_voices(svpath(s, VSCHEME)->vscheme);
		    for (vr_p = svr_p->vnolist_p; vr_p != 0; vr_p = vr_p->next) {
			for (v = vr_p->begin; v <= vr_p->end && v <= numvoices; v++) {
				if (clone_src_p->staffno == s
						&& clone_src_p->voiceno == v) {
					/* skip past the clone source */
					continue;
				}
				clone1ssv(clone_src_p, insert_p, s, v);
			}
		    }
		}
	    }
	}
}

/* Clone the given SSV and insert it in the specified place in the main list.
 * Change the copied versions staffno/voiceno to the specified values,
 * and do any additional error checks and processsing as needed.
 */

static void
clone1ssv(clone_src_p, insert_p, staffno, voiceno)

struct SSV *clone_src_p;
struct MAINLL *insert_p;
int staffno;
int voiceno;

{
	struct MAINLL *newmllssv_p;	/* the cloned SSV */

	newmllssv_p = newMAINLLstruct(S_SSV, Currstruct_p->inputlineno);
	memcpy(newmllssv_p->u.ssv_p, clone_src_p, sizeof(struct SSV));
	newmllssv_p->u.ssv_p->staffno = staffno;
	newmllssv_p->u.ssv_p->voiceno = voiceno;
	/* Fix beamstyle/timeunit. It could be different for each clone */
	remember_tsig_params(newmllssv_p);
	insertMAINLL(newmllssv_p, insert_p);

	/* Make sure any tab staffs have non-tabs above them.
	 * One disadvantage of waiting till making the clone to check this is
	 * that if there is a user error, the line number will be the end
	 * of the context rather than the one with "stafflines" on it.
	 * But the error message will tell them which staff is the problem,
	 * and this should be pretty rare--they have to be setting stafflines
	 * in a staff context for multiple staffs where the combination of
	 * staffs is illegal--so it doesn't seem worth the extra work it would
	 * be to be more precise. */
	if (newmllssv_p->u.ssv_p->context == C_STAFF &&
				newmllssv_p->u.ssv_p->used[STAFFLINES] == YES) {
		chk_tab(newmllssv_p);
	}
	asgnssv(newmllssv_p->u.ssv_p);
}


/* Check that the staffs/voices specified for an SSV are valid.
 * We only give a warning if all values specified are out of range.
 * If at least one is valid, we assume user did something like
 * 	voice 1-40 3
 * to set something for voice 3 on all staffs without having to worry
 * about which staffs actually have a voice 3.
 */

static void
chk_ssv_ranges(context)

int context;	/* expected to only be either C_STAFF or C_VOICE */

{
	struct SVRANGELIST *svr_p;	/* list of staff/voice ranges */
	struct RANGELIST *sr_p;		/* list of staff ranges */
	struct RANGELIST *vr_p;		/* list of voice ranges */
	int s;				/* staff number */
	int v;				/* voice number */
	int numvoices;			/* how many voices on current staff */


	for (svr_p = Svrangelist_p; svr_p != 0; svr_p = svr_p->next) {
	    for (sr_p = svr_p->stafflist_p; sr_p != 0; sr_p = sr_p->next) { 
		for (s = sr_p->begin; s <= sr_p->end; s++) {
		    if (context == C_STAFF) {
			if (s >= MINSTAFFS && s <= Score.staffs) {
				/* Found a valid one. One is all we need. */
				return;
			}
			continue;
		    }

		    /* a voice context */
		    numvoices = vscheme_voices(svpath(s, VSCHEME)->vscheme);
		    for (vr_p = svr_p->vnolist_p; vr_p != 0; vr_p = vr_p->next) { 
			for (v = vr_p->begin; v <= vr_p->end; v++) {
			    if (s >= MINSTAFFS && s <= Score.staffs &&
					v >= MINVOICES && v <= numvoices) {
				/* Found a valid one. One is all we need. */
				return;
			    }
			}
		    }
		}
	    }
	}
	/* Fell through all the loops without finding a single valid case */
	if (context == C_STAFF) {
		l_warning(Curr_filename, yylineno,
			"staff out of range; must be between %d and %d",
			MINSTAFFS, Score.staffs);
	}
	else {
		l_warning(Curr_filename, yylineno,
			"staff/voice combination out of range; staff must be between %d and %d, voice between %d and %d and within vscheme limit for the staff",
			MINSTAFFS, MAXSTAFFS, MINVOICES, MAXVOICES);
	}
}


/* if defining more than one voice at the moment, not valid to set
 * a location variable. Print message and reset flag so we don't
 * give any more of the same message this measure */

static void
var_valid()

{
	if (Defining_multiple == YES) {
		yyerror("location tag is only allowed when defining a single voice");
		Defining_multiple = NO;
	}
	else if (Chord_at_a_time == YES) {
		yyerror("location tag not allowed with chord-at-a-time input");
	}
}


/* do a print command like left, right, or center. Allocate a PRINTDATA
 * struct and fill in all the information about what to print and where
 * to print it. */

static void
proc_printcmd(justifytype, inpc_p, str, font, size, got_nl, isPostScript, extra)

int justifytype;	/* J_LEFT, etc */
struct INPCOORD *inpc_p;	/* where to print */
char *str;		/* the string to print */
int font;
int size;
int got_nl;	/* YES if should go to next line before printing */
int isPostScript;	/* YES if is raw PostScript rather than normal print */
double extra;	/* how much extra vetical padding to add */

{
	struct PRINTDATA *curr_print_p;


	(void) contextcheck(C_MUSIC | C_BLOCKHEAD, "print");

	/* save all the info in a PRINTDATA struct */
	MALLOC(PRINTDATA, curr_print_p, 1);

	if (inpc_p == (struct INPCOORD *) 0) {

		/* print at default location */
		curr_print_p->location.vert_p = symval("_cur", (float **) 0);

		switch (justifytype) {
		case J_LEFT:
		case J_RAGPARA:
		case J_JUSTPARA:
			/* default x is relative to _win.w */
			curr_print_p->location.hor_p = symval("_win", (float **) 0);
			curr_print_p->location.htype = AW;
			break;
		case J_RIGHT:
			/* default x is relative to _win.e */
			curr_print_p->location.hor_p = symval("_win", (float **) 0);
			curr_print_p->location.htype = AE;
			break;
		case J_CENTER:
			/* default x is relative to _win.x */
			curr_print_p->location.hor_p = symval("_win", (float **) 0);
			curr_print_p->location.htype = AX;
			break;
		case J_NONE:
		default:
			/* default is at current location */
			curr_print_p->location.hor_p = symval("_cur", (float **) 0);
			curr_print_p->location.htype = AX;
			break;
		}

		curr_print_p->location.hsteps = 0.0;
		curr_print_p->location.counts = 0.0;
		curr_print_p->location.vtype = AY;
		curr_print_p->location.vsteps = -extra;

		/* If this is the first print command in a block,
		 * we implicitly go down by the font ascent. */
		if (isPostScript == NO && Currblock_p != 0
					&& Next_print_link_p_p
					== &(Currblock_p->printdata_p) ) {
			curr_print_p->location.vsteps -= fontascent(font, size)
						/ STEPSIZE;
			got_nl = NO;
		}

	}
	else {
		curr_print_p->location = *(inpc_p);
		rep_inpcoord(inpc_p, &(curr_print_p->location) );
		FREE(inpc_p);
	}

	curr_print_p->isPostScript = isPostScript;
	if (isPostScript == YES) {
		curr_print_p->width = 0.0;
	}
	else {
		curr_print_p->width = strwidth(str);
	}
	curr_print_p->justifytype = (short) justifytype;
	curr_print_p->string = str;
	curr_print_p->inputfile = Curr_filename;
	curr_print_p->inputlineno = (short) yylineno;

	/* special case of user asking that this be printed
	 * on the next line. In this case we go down by the
	 * current point size */
	if (got_nl) {
		curr_print_p->location.vsteps -=
					fontheight(font, size) / STEPSIZE;
	}

	/* Now link onto proper list -- could be in BLOCKHEAD
	 * context or doing a PRHEAD. If we are in BLOCKHEAD context, we
	 * definitely have something in the main list to attach to.
	 * If in C_MUSIC, we may or may not depending on whether
	 * this is the first in a series of print commands. */
	if ( ((Context & C_BLOCKHEAD) != 0)
		|| ((Currstruct_p != (struct MAINLL *) 0) &&
		(Currstruct_p->str == S_PRHEAD) ) ) {
			if (Next_print_link_p_p == (struct PRINTDATA **) 0) {
				pfatal("Next_print_link_p_p is null");
			}
			/* this means the value of Next_print_link_p_p
			 * is valid and will tell us where to link */
			*Next_print_link_p_p = curr_print_p;
	}
	else {
		/* We must be in music context, but the last
		 * thing we saw wasn't a print command, so
		 * we have to allocate a PRHEAD and put it
		 * in the main list and attach to that. */
		/* allocate a new struct, put on main list and attach
		 * the print command to it */
		Currstruct_p = newMAINLLstruct(S_PRHEAD, yylineno);
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p->u.prhead_p->printdata_p = curr_print_p;
		Currstruct_p = (struct MAINLL *) 0;
	}

	/* in any case, if we get another print command right away, it
	 * should be linked onto this one, so remember where we are */
	Next_print_link_p_p = &(curr_print_p->next);
	curr_print_p->next = (struct PRINTDATA *) 0;
}


/* if the given string has a higher ascent than the given font/size, return
 * the amount is it higher, in stepsizes */

static double
extra_needed(font, size, string)

int font;	/* default font for this string */
int size;	/* default size for this string */
char *string;	/* string, which might contain \s(xx) to make it bigger
		 * than normal */

{
	double usual_ascent;
	double this_string_ascent;


	/* get expected ascent for this font/size and actual for string */
	usual_ascent = fontascent(font, size);
	(void) fix_string(string, font, size, Curr_filename, yylineno);
	this_string_ascent = strascent(string);

	/* if this string is too tall, return by how much it is too tall */
	if (this_string_ascent > usual_ascent) {
		return (this_string_ascent - usual_ascent) / STEPSIZE;
	}
	else {
		return(0.0);
	}
}


/* Sometimes lexer will recognize something as a keyword, but it's
 * really a bunch of note letters, due to context. This function turns
 * the mistaken keyword into notes. It handles a through g and r and s, and
 * us, but not n, which would also be troublesome, but which we don't (yet)
 * need to handle.
 */

static void
keyword_notes(str)

char *str;

{
	for (  ; *str != '\0'; str++) {
		if (*str == 'u' && *(str+1) == 's') {
			str++;
			continue;
		}

		/* On the last letter, if it's a true pitch, have to
		 * push it back into input, because it could be followed
		 * by an accidental or octave mark or tick or something. */
		if (*(str+1) == '\0' && *str >= 'a' && *str <= 'g') {
			pushback(*str);
		}
		else {
			add_note(Curr_grpsyl_p, (int) *str, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
		}
	}
}


/* Free up the current additive time list, if any */

static void
free_extra_time()
{
	struct TIMELIST *tl_p;

	/* If this was the most recently allocated list, we can free it.
	 * We can't free otherwise, since Extra_time_p might be
	 * pointing to one from an SSV, which must be kept around. */
	if (Extra_time_p == Last_alloced_timelist_p) {
		while (Extra_time_p != 0) {
			tl_p = Extra_time_p->next;
			FREE(Extra_time_p);
			Extra_time_p = tl_p;
		}
	}
	Extra_time_p = 0;
	Last_alloced_timelist_p = 0;
	Curr_timelist_p = 0;
}


/* Add an item to the internal representation of the time signature
 * currently being collected */

static void
tsig_item(item)

int item;	/* the byte to add to timerep */

{
	if (Tsig_offset >= MAXTSLEN  - 1) {
		l_ufatal(Curr_filename, yylineno,
				"time signature is too complicated");
	}
	Timerep[Tsig_offset++] = (char) item;
}
